from okareo.checks import CheckResponse, CodeBasedCheck


class Check(CodeBasedCheck):
    @staticmethod
    def evaluate(model_output: str, scenario_input: str, scenario_result: str, metadata: dict, model_input: str) -> CheckResponse:  # fmt: skip
        return (
            CheckResponse(
                score=True, explanation="The combined length is at least 20 characters."
            )
            if len(model_output + scenario_input + scenario_result + str(model_input))
            + len(metadata)
            >= 20
            else CheckResponse(
                score=False,
                explanation="The combined length is less than 20 characters.",
            )
        )



""" Contains all the data models used in inputs/outputs """

from .add_model_to_group_v0_groups_group_id_models_post_response_add_model_to_group_v0_groups_group_id_models_post import (
    AddModelToGroupV0GroupsGroupIdModelsPostResponseAddModelToGroupV0GroupsGroupIdModelsPost,
)
from .assistant_end_thread_request import AssistantEndThreadRequest
from .assistant_end_thread_response import AssistantEndThreadResponse
from .assistant_message_request import AssistantMessageRequest
from .assistant_message_response import AssistantMessageResponse
from .assistant_thread_response import AssistantThreadResponse
from .body_check_delete_v0_check_check_id_delete import BodyCheckDeleteV0CheckCheckIdDelete
from .body_check_upload_v0_check_upload_post import BodyCheckUploadV0CheckUploadPost
from .body_scenario_sets_upload_v0_scenario_sets_upload_post import BodyScenarioSetsUploadV0ScenarioSetsUploadPost
from .check_create_update_schema import CheckCreateUpdateSchema
from .check_create_update_schema_check_config import CheckCreateUpdateSchemaCheckConfig
from .comparison_operator import ComparisonOperator
from .create_group_v0_groups_post_response_create_group_v0_groups_post import (
    CreateGroupV0GroupsPostResponseCreateGroupV0GroupsPost,
)
from .create_group_v0_groups_post_source import CreateGroupV0GroupsPostSource
from .create_trace_eval_v0_groups_group_id_trace_eval_post_response_create_trace_eval_v0_groups_group_id_trace_eval_post import (
    CreateTraceEvalV0GroupsGroupIdTraceEvalPostResponseCreateTraceEvalV0GroupsGroupIdTraceEvalPost,
)
from .datapoint_field import DatapointField
from .datapoint_filter_create import DatapointFilterCreate
from .datapoint_filter_delete import DatapointFilterDelete
from .datapoint_filter_item import DatapointFilterItem
from .datapoint_filter_item_average_metrics_type_0 import DatapointFilterItemAverageMetricsType0
from .datapoint_filter_item_latest_test_run_type_0 import DatapointFilterItemLatestTestRunType0
from .datapoint_filter_search import DatapointFilterSearch
from .datapoint_filter_update import DatapointFilterUpdate
from .datapoint_list_item import DatapointListItem
from .datapoint_list_item_agent_metadata_type_0 import DatapointListItemAgentMetadataType0
from .datapoint_list_item_checks import DatapointListItemChecks
from .datapoint_list_item_input_tools_item import DatapointListItemInputToolsItem
from .datapoint_list_item_model_metadata_type_0 import DatapointListItemModelMetadataType0
from .datapoint_list_item_result_embeddings_item import DatapointListItemResultEmbeddingsItem
from .datapoint_list_item_result_tool_calls_item import DatapointListItemResultToolCallsItem
from .datapoint_list_item_user_metadata_type_0 import DatapointListItemUserMetadataType0
from .datapoint_response import DatapointResponse
from .datapoint_schema import DatapointSchema
from .datapoint_search import DatapointSearch
from .datapoint_summary_item import DatapointSummaryItem
from .datapoint_summary_item_user_metadata_type_0 import DatapointSummaryItemUserMetadataType0
from .datapoint_tags_schema import DatapointTagsSchema
from .error_response import ErrorResponse
from .evaluation_payload import EvaluationPayload
from .evaluation_payload_metrics_kwargs import EvaluationPayloadMetricsKwargs
from .evaluator_brief_response import EvaluatorBriefResponse
from .evaluator_brief_response_check_config import EvaluatorBriefResponseCheckConfig
from .evaluator_detailed_response import EvaluatorDetailedResponse
from .evaluator_detailed_response_check_config import EvaluatorDetailedResponseCheckConfig
from .evaluator_generate_response import EvaluatorGenerateResponse
from .evaluator_spec_request import EvaluatorSpecRequest
from .feedback_range_summary import FeedbackRangeSummary
from .filter_condition import FilterCondition
from .find_test_data_point_payload import FindTestDataPointPayload
from .full_data_point_item import FullDataPointItem
from .full_data_point_item_metric_value import FullDataPointItemMetricValue
from .full_data_point_item_model_metadata_type_0 import FullDataPointItemModelMetadataType0
from .full_data_point_item_scenario_input_type_0 import FullDataPointItemScenarioInputType0
from .full_data_point_item_scenario_result_type_0 import FullDataPointItemScenarioResultType0
from .general_find_payload import GeneralFindPayload
from .generate_slack_auth_url_v0_generate_slack_auth_url_post_project_data import (
    GenerateSlackAuthUrlV0GenerateSlackAuthUrlPostProjectData,
)
from .generate_slack_auth_url_v0_generate_slack_auth_url_post_response_generate_slack_auth_url_v0_generate_slack_auth_url_post import (
    GenerateSlackAuthUrlV0GenerateSlackAuthUrlPostResponseGenerateSlackAuthUrlV0GenerateSlackAuthUrlPost,
)
from .generation_tone import GenerationTone
from .get_datapoint_counts_v0_filter_counts_post_response_get_datapoint_counts_v0_filter_counts_post import (
    GetDatapointCountsV0FilterCountsPostResponseGetDatapointCountsV0FilterCountsPost,
)
from .get_filter_group_notifications_v0_filter_group_notifications_get_response_200_item import (
    GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item,
)
from .get_groups_v0_groups_get_response_200_item import GetGroupsV0GroupsGetResponse200Item
from .get_notification_history_v0_notification_history_get_response_200_item import (
    GetNotificationHistoryV0NotificationHistoryGetResponse200Item,
)
from .group_schema import GroupSchema
from .http_validation_error import HTTPValidationError
from .model_under_test_response import ModelUnderTestResponse
from .model_under_test_response_models import ModelUnderTestResponseModels
from .model_under_test_response_models_additional_property import ModelUnderTestResponseModelsAdditionalProperty
from .model_under_test_schema import ModelUnderTestSchema
from .model_under_test_schema_models import ModelUnderTestSchemaModels
from .model_under_test_schema_models_additional_property import ModelUnderTestSchemaModelsAdditionalProperty
from .project_response import ProjectResponse
from .project_schema import ProjectSchema
from .scenario_data_poin_response import ScenarioDataPoinResponse
from .scenario_data_poin_response_meta_data import ScenarioDataPoinResponseMetaData
from .scenario_set_create import ScenarioSetCreate
from .scenario_set_generate import ScenarioSetGenerate
from .scenario_set_generate_checks_item_type_1 import ScenarioSetGenerateChecksItemType1
from .scenario_set_generate_generation_schema import ScenarioSetGenerateGenerationSchema
from .scenario_set_response import ScenarioSetResponse
from .scenario_set_update import ScenarioSetUpdate
from .scenario_type import ScenarioType
from .seed_data import SeedData
from .setup_filter_group_notification_v0_setup_filter_group_notification_post_response_setup_filter_group_notification_v0_setup_filter_group_notification_post import (
    SetupFilterGroupNotificationV0SetupFilterGroupNotificationPostResponseSetupFilterGroupNotificationV0SetupFilterGroupNotificationPost,
)
from .slack_oauth_callback_v0_slack_get_response_slack_oauth_callback_v0_slack_get import (
    SlackOauthCallbackV0SlackGetResponseSlackOauthCallbackV0SlackGet,
)
from .summary_datapoint_search import SummaryDatapointSearch
from .test_custom_endpoint_request import TestCustomEndpointRequest
from .test_custom_endpoint_request_end_session_params import TestCustomEndpointRequestEndSessionParams
from .test_custom_endpoint_request_next_message_params import TestCustomEndpointRequestNextMessageParams
from .test_custom_endpoint_request_start_session_params import TestCustomEndpointRequestStartSessionParams
from .test_custom_endpoint_response import TestCustomEndpointResponse
from .test_custom_endpoint_response_end_session_raw_response import TestCustomEndpointResponseEndSessionRawResponse
from .test_custom_endpoint_response_next_message_raw_response import TestCustomEndpointResponseNextMessageRawResponse
from .test_custom_endpoint_response_start_session_raw_response import TestCustomEndpointResponseStartSessionRawResponse
from .test_data_point_item import TestDataPointItem
from .test_data_point_item_metric_value import TestDataPointItemMetricValue
from .test_data_point_payload import TestDataPointPayload
from .test_data_point_response import TestDataPointResponse
from .test_run_item import TestRunItem
from .test_run_item_model_metrics import TestRunItemModelMetrics
from .test_run_payload import TestRunPayload
from .test_run_payload_v2 import TestRunPayloadV2
from .test_run_payload_v2_api_keys import TestRunPayloadV2ApiKeys
from .test_run_payload_v2_metrics_kwargs import TestRunPayloadV2MetricsKwargs
from .test_run_payload_v2_model_results import TestRunPayloadV2ModelResults
from .test_run_type import TestRunType
from .update_test_data_point_payload import UpdateTestDataPointPayload
from .validation_error import ValidationError

__all__ = (
    "AddModelToGroupV0GroupsGroupIdModelsPostResponseAddModelToGroupV0GroupsGroupIdModelsPost",
    "AssistantEndThreadRequest",
    "AssistantEndThreadResponse",
    "AssistantMessageRequest",
    "AssistantMessageResponse",
    "AssistantThreadResponse",
    "BodyCheckDeleteV0CheckCheckIdDelete",
    "BodyCheckUploadV0CheckUploadPost",
    "BodyScenarioSetsUploadV0ScenarioSetsUploadPost",
    "CheckCreateUpdateSchema",
    "CheckCreateUpdateSchemaCheckConfig",
    "ComparisonOperator",
    "CreateGroupV0GroupsPostResponseCreateGroupV0GroupsPost",
    "CreateGroupV0GroupsPostSource",
    "CreateTraceEvalV0GroupsGroupIdTraceEvalPostResponseCreateTraceEvalV0GroupsGroupIdTraceEvalPost",
    "DatapointField",
    "DatapointFilterCreate",
    "DatapointFilterDelete",
    "DatapointFilterItem",
    "DatapointFilterItemAverageMetricsType0",
    "DatapointFilterItemLatestTestRunType0",
    "DatapointFilterSearch",
    "DatapointFilterUpdate",
    "DatapointListItem",
    "DatapointListItemAgentMetadataType0",
    "DatapointListItemChecks",
    "DatapointListItemInputToolsItem",
    "DatapointListItemModelMetadataType0",
    "DatapointListItemResultEmbeddingsItem",
    "DatapointListItemResultToolCallsItem",
    "DatapointListItemUserMetadataType0",
    "DatapointResponse",
    "DatapointSchema",
    "DatapointSearch",
    "DatapointSummaryItem",
    "DatapointSummaryItemUserMetadataType0",
    "DatapointTagsSchema",
    "ErrorResponse",
    "EvaluationPayload",
    "EvaluationPayloadMetricsKwargs",
    "EvaluatorBriefResponse",
    "EvaluatorBriefResponseCheckConfig",
    "EvaluatorDetailedResponse",
    "EvaluatorDetailedResponseCheckConfig",
    "EvaluatorGenerateResponse",
    "EvaluatorSpecRequest",
    "FeedbackRangeSummary",
    "FilterCondition",
    "FindTestDataPointPayload",
    "FullDataPointItem",
    "FullDataPointItemMetricValue",
    "FullDataPointItemModelMetadataType0",
    "FullDataPointItemScenarioInputType0",
    "FullDataPointItemScenarioResultType0",
    "GeneralFindPayload",
    "GenerateSlackAuthUrlV0GenerateSlackAuthUrlPostProjectData",
    "GenerateSlackAuthUrlV0GenerateSlackAuthUrlPostResponseGenerateSlackAuthUrlV0GenerateSlackAuthUrlPost",
    "GenerationTone",
    "GetDatapointCountsV0FilterCountsPostResponseGetDatapointCountsV0FilterCountsPost",
    "GetFilterGroupNotificationsV0FilterGroupNotificationsGetResponse200Item",
    "GetGroupsV0GroupsGetResponse200Item",
    "GetNotificationHistoryV0NotificationHistoryGetResponse200Item",
    "GroupSchema",
    "HTTPValidationError",
    "ModelUnderTestResponse",
    "ModelUnderTestResponseModels",
    "ModelUnderTestResponseModelsAdditionalProperty",
    "ModelUnderTestSchema",
    "ModelUnderTestSchemaModels",
    "ModelUnderTestSchemaModelsAdditionalProperty",
    "ProjectResponse",
    "ProjectSchema",
    "ScenarioDataPoinResponse",
    "ScenarioDataPoinResponseMetaData",
    "ScenarioSetCreate",
    "ScenarioSetGenerate",
    "ScenarioSetGenerateChecksItemType1",
    "ScenarioSetGenerateGenerationSchema",
    "ScenarioSetResponse",
    "ScenarioSetUpdate",
    "ScenarioType",
    "SeedData",
    "SetupFilterGroupNotificationV0SetupFilterGroupNotificationPostResponseSetupFilterGroupNotificationV0SetupFilterGroupNotificationPost",
    "SlackOauthCallbackV0SlackGetResponseSlackOauthCallbackV0SlackGet",
    "SummaryDatapointSearch",
    "TestCustomEndpointRequest",
    "TestCustomEndpointRequestEndSessionParams",
    "TestCustomEndpointRequestNextMessageParams",
    "TestCustomEndpointRequestStartSessionParams",
    "TestCustomEndpointResponse",
    "TestCustomEndpointResponseEndSessionRawResponse",
    "TestCustomEndpointResponseNextMessageRawResponse",
    "TestCustomEndpointResponseStartSessionRawResponse",
    "TestDataPointItem",
    "TestDataPointItemMetricValue",
    "TestDataPointPayload",
    "TestDataPointResponse",
    "TestRunItem",
    "TestRunItemModelMetrics",
    "TestRunPayload",
    "TestRunPayloadV2",
    "TestRunPayloadV2ApiKeys",
    "TestRunPayloadV2MetricsKwargs",
    "TestRunPayloadV2ModelResults",
    "TestRunType",
    "UpdateTestDataPointPayload",
    "ValidationError",
)



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.datapoint_filter_item_average_metrics_type_0 import DatapointFilterItemAverageMetricsType0
    from ..models.datapoint_filter_item_latest_test_run_type_0 import DatapointFilterItemLatestTestRunType0
    from ..models.filter_condition import FilterCondition


T = TypeVar("T", bound="DatapointFilterItem")


@_attrs_define
class DatapointFilterItem:
    """
    Attributes:
        filters (List['FilterCondition']): List of filter conditions to apply
        name (Union[Unset, str]): Optional name describing this filter
        description (Union[Unset, str]): Optional description for this filter
        checks (Union[Unset, List[str]]): Checks to apply to datapoints in the filter.
        project_id (Union[Unset, str]): Project ID these filters belong to
        filter_group_id (Union[Unset, str]): Group ID for filter
        latest_test_run (Union['DatapointFilterItemLatestTestRunType0', Any, Unset]): Group ID for filter
        datapoint_count (Union[Any, Unset, int]): Group ID for filter
        issue_count (Union[Any, Unset, int]): Count of issues for this filter
        error_count (Union[Any, Unset, int]): Count of errors for this filter
        average_metrics (Union['DatapointFilterItemAverageMetricsType0', Any, Unset]): Metrics for checks in this filter
        failed_checks (Union[Unset, Any]): Array of failed cheßck names
        avg_latency (Union[Any, Unset, float]): Average latency for completions
        total_cost (Union[Any, Unset, float]): Total dollar cost of all completions
        avg_num_turns (Union[Any, Unset, float]): Average number of turns for completions
        slack_enabled (Union[Unset, bool]): Whether Slack notifications are enabled for this filter group
        email_enabled (Union[Unset, bool]): Whether Email notifications are enabled for this filter group
        time_created (Union[Unset, datetime.datetime]): Created datetime for this filter group
        time_updated (Union[Unset, datetime.datetime]): Last updated datetime for this filter group
    """

    filters: List["FilterCondition"]
    name: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    checks: Union[Unset, List[str]] = UNSET
    project_id: Union[Unset, str] = UNSET
    filter_group_id: Union[Unset, str] = UNSET
    latest_test_run: Union["DatapointFilterItemLatestTestRunType0", Any, Unset] = UNSET
    datapoint_count: Union[Any, Unset, int] = UNSET
    issue_count: Union[Any, Unset, int] = UNSET
    error_count: Union[Any, Unset, int] = UNSET
    average_metrics: Union["DatapointFilterItemAverageMetricsType0", Any, Unset] = UNSET
    failed_checks: Union[Unset, Any] = UNSET
    avg_latency: Union[Any, Unset, float] = UNSET
    total_cost: Union[Any, Unset, float] = UNSET
    avg_num_turns: Union[Any, Unset, float] = UNSET
    slack_enabled: Union[Unset, bool] = UNSET
    email_enabled: Union[Unset, bool] = UNSET
    time_created: Union[Unset, datetime.datetime] = UNSET
    time_updated: Union[Unset, datetime.datetime] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.datapoint_filter_item_average_metrics_type_0 import DatapointFilterItemAverageMetricsType0
        from ..models.datapoint_filter_item_latest_test_run_type_0 import DatapointFilterItemLatestTestRunType0

        filters = []
        for filters_item_data in self.filters:
            filters_item = filters_item_data.to_dict()

            filters.append(filters_item)

        name = self.name
        description = self.description
        checks: Union[Unset, List[str]] = UNSET
        if not isinstance(self.checks, Unset):
            checks = self.checks

        project_id = self.project_id
        filter_group_id = self.filter_group_id
        latest_test_run: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.latest_test_run, Unset):
            latest_test_run = UNSET

        elif isinstance(self.latest_test_run, DatapointFilterItemLatestTestRunType0):
            latest_test_run = UNSET
            if not isinstance(self.latest_test_run, Unset):
                latest_test_run = self.latest_test_run.to_dict()

        else:
            latest_test_run = self.latest_test_run

        datapoint_count: Union[Any, Unset, int]
        if isinstance(self.datapoint_count, Unset):
            datapoint_count = UNSET

        else:
            datapoint_count = self.datapoint_count

        issue_count: Union[Any, Unset, int]
        if isinstance(self.issue_count, Unset):
            issue_count = UNSET

        else:
            issue_count = self.issue_count

        error_count: Union[Any, Unset, int]
        if isinstance(self.error_count, Unset):
            error_count = UNSET

        else:
            error_count = self.error_count

        average_metrics: Union[Any, Dict[str, Any], Unset]
        if isinstance(self.average_metrics, Unset):
            average_metrics = UNSET

        elif isinstance(self.average_metrics, DatapointFilterItemAverageMetricsType0):
            average_metrics = UNSET
            if not isinstance(self.average_metrics, Unset):
                average_metrics = self.average_metrics.to_dict()

        else:
            average_metrics = self.average_metrics

        failed_checks = self.failed_checks
        avg_latency: Union[Any, Unset, float]
        if isinstance(self.avg_latency, Unset):
            avg_latency = UNSET

        else:
            avg_latency = self.avg_latency

        total_cost: Union[Any, Unset, float]
        if isinstance(self.total_cost, Unset):
            total_cost = UNSET

        else:
            total_cost = self.total_cost

        avg_num_turns: Union[Any, Unset, float]
        if isinstance(self.avg_num_turns, Unset):
            avg_num_turns = UNSET

        else:
            avg_num_turns = self.avg_num_turns

        slack_enabled = self.slack_enabled
        email_enabled = self.email_enabled
        time_created: Union[Unset, str] = UNSET
        if not isinstance(self.time_created, Unset):
            time_created = self.time_created.isoformat()

        time_updated: Union[Unset, str] = UNSET
        if not isinstance(self.time_updated, Unset):
            time_updated = self.time_updated.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "filters": filters,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if description is not UNSET:
            field_dict["description"] = description
        if checks is not UNSET:
            field_dict["checks"] = checks
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if filter_group_id is not UNSET:
            field_dict["filter_group_id"] = filter_group_id
        if latest_test_run is not UNSET:
            field_dict["latest_test_run"] = latest_test_run
        if datapoint_count is not UNSET:
            field_dict["datapoint_count"] = datapoint_count
        if issue_count is not UNSET:
            field_dict["issue_count"] = issue_count
        if error_count is not UNSET:
            field_dict["error_count"] = error_count
        if average_metrics is not UNSET:
            field_dict["average_metrics"] = average_metrics
        if failed_checks is not UNSET:
            field_dict["failed_checks"] = failed_checks
        if avg_latency is not UNSET:
            field_dict["avg_latency"] = avg_latency
        if total_cost is not UNSET:
            field_dict["total_cost"] = total_cost
        if avg_num_turns is not UNSET:
            field_dict["avg_num_turns"] = avg_num_turns
        if slack_enabled is not UNSET:
            field_dict["slack_enabled"] = slack_enabled
        if email_enabled is not UNSET:
            field_dict["email_enabled"] = email_enabled
        if time_created is not UNSET:
            field_dict["time_created"] = time_created
        if time_updated is not UNSET:
            field_dict["time_updated"] = time_updated

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.datapoint_filter_item_average_metrics_type_0 import DatapointFilterItemAverageMetricsType0
        from ..models.datapoint_filter_item_latest_test_run_type_0 import DatapointFilterItemLatestTestRunType0
        from ..models.filter_condition import FilterCondition

        d = src_dict.copy()
        filters = []
        _filters = d.pop("filters")
        for filters_item_data in _filters:
            filters_item = FilterCondition.from_dict(filters_item_data)

            filters.append(filters_item)

        name = d.pop("name", UNSET)

        description = d.pop("description", UNSET)

        checks = cast(List[str], d.pop("checks", UNSET))

        project_id = d.pop("project_id", UNSET)

        filter_group_id = d.pop("filter_group_id", UNSET)

        def _parse_latest_test_run(data: object) -> Union["DatapointFilterItemLatestTestRunType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _latest_test_run_type_0 = data
                latest_test_run_type_0: Union[Unset, DatapointFilterItemLatestTestRunType0]
                if isinstance(_latest_test_run_type_0, Unset):
                    latest_test_run_type_0 = UNSET
                else:
                    latest_test_run_type_0 = DatapointFilterItemLatestTestRunType0.from_dict(_latest_test_run_type_0)

                return latest_test_run_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointFilterItemLatestTestRunType0", Any, Unset], data)

        latest_test_run = _parse_latest_test_run(d.pop("latest_test_run", UNSET))

        def _parse_datapoint_count(data: object) -> Union[Any, Unset, int]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, int], data)

        datapoint_count = _parse_datapoint_count(d.pop("datapoint_count", UNSET))

        def _parse_issue_count(data: object) -> Union[Any, Unset, int]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, int], data)

        issue_count = _parse_issue_count(d.pop("issue_count", UNSET))

        def _parse_error_count(data: object) -> Union[Any, Unset, int]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, int], data)

        error_count = _parse_error_count(d.pop("error_count", UNSET))

        def _parse_average_metrics(data: object) -> Union["DatapointFilterItemAverageMetricsType0", Any, Unset]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                _average_metrics_type_0 = data
                average_metrics_type_0: Union[Unset, DatapointFilterItemAverageMetricsType0]
                if isinstance(_average_metrics_type_0, Unset):
                    average_metrics_type_0 = UNSET
                else:
                    average_metrics_type_0 = DatapointFilterItemAverageMetricsType0.from_dict(_average_metrics_type_0)

                return average_metrics_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatapointFilterItemAverageMetricsType0", Any, Unset], data)

        average_metrics = _parse_average_metrics(d.pop("average_metrics", UNSET))

        failed_checks = d.pop("failed_checks", UNSET)

        def _parse_avg_latency(data: object) -> Union[Any, Unset, float]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, float], data)

        avg_latency = _parse_avg_latency(d.pop("avg_latency", UNSET))

        def _parse_total_cost(data: object) -> Union[Any, Unset, float]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, float], data)

        total_cost = _parse_total_cost(d.pop("total_cost", UNSET))

        def _parse_avg_num_turns(data: object) -> Union[Any, Unset, float]:
            if isinstance(data, Unset):
                return data
            return cast(Union[Any, Unset, float], data)

        avg_num_turns = _parse_avg_num_turns(d.pop("avg_num_turns", UNSET))

        slack_enabled = d.pop("slack_enabled", UNSET)

        email_enabled = d.pop("email_enabled", UNSET)

        _time_created = d.pop("time_created", UNSET)
        time_created: Union[Unset, datetime.datetime]
        if isinstance(_time_created, Unset):
            time_created = UNSET
        else:
            time_created = isoparse(_time_created)

        _time_updated = d.pop("time_updated", UNSET)
        time_updated: Union[Unset, datetime.datetime]
        if isinstance(_time_updated, Unset):
            time_updated = UNSET
        else:
            time_updated = isoparse(_time_updated)

        datapoint_filter_item = cls(
            filters=filters,
            name=name,
            description=description,
            checks=checks,
            project_id=project_id,
            filter_group_id=filter_group_id,
            latest_test_run=latest_test_run,
            datapoint_count=datapoint_count,
            issue_count=issue_count,
            error_count=error_count,
            average_metrics=average_metrics,
            failed_checks=failed_checks,
            avg_latency=avg_latency,
            total_cost=total_cost,
            avg_num_turns=avg_num_turns,
            slack_enabled=slack_enabled,
            email_enabled=email_enabled,
            time_created=time_created,
            time_updated=time_updated,
        )

        datapoint_filter_item.additional_properties = d
        return datapoint_filter_item

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from io import BytesIO
from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, File, FileJsonType, Unset

T = TypeVar("T", bound="BodyCheckUploadV0CheckUploadPost")


@_attrs_define
class BodyCheckUploadV0CheckUploadPost:
    """
    Attributes:
        name (str): Name of the Check
        requires_scenario_input (bool): Whether the check requires scenario input
        requires_scenario_result (bool): Whether the check requires scenario expected result
        description (Union[Unset, str]): Description of the Check Default: 'No description provided'.
        output_data_type (Union[Unset, str]): Check output data type (i.e., bool, int, float)
        project_id (Union[Unset, str]): ID for the project
        file (Union[Unset, File]):
        update (Union[Unset, bool]): Update the check
    """

    name: str
    requires_scenario_input: bool
    requires_scenario_result: bool
    description: Union[Unset, str] = "No description provided"
    output_data_type: Union[Unset, str] = UNSET
    project_id: Union[Unset, str] = UNSET
    file: Union[Unset, File] = UNSET
    update: Union[Unset, bool] = False
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name
        requires_scenario_input = self.requires_scenario_input
        requires_scenario_result = self.requires_scenario_result
        description = self.description
        output_data_type = self.output_data_type
        project_id = self.project_id
        file: Union[Unset, FileJsonType] = UNSET
        if not isinstance(self.file, Unset):
            file = self.file.to_tuple()

        update = self.update

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "requires_scenario_input": requires_scenario_input,
                "requires_scenario_result": requires_scenario_result,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if output_data_type is not UNSET:
            field_dict["output_data_type"] = output_data_type
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if file is not UNSET:
            field_dict["file"] = file
        if update is not UNSET:
            field_dict["update"] = update

        return field_dict

    def to_multipart(self) -> Dict[str, Any]:
        name = self.name if isinstance(self.name, Unset) else (None, str(self.name).encode(), "text/plain")
        requires_scenario_input = (
            self.requires_scenario_input
            if isinstance(self.requires_scenario_input, Unset)
            else (None, str(self.requires_scenario_input).encode(), "text/plain")
        )
        requires_scenario_result = (
            self.requires_scenario_result
            if isinstance(self.requires_scenario_result, Unset)
            else (None, str(self.requires_scenario_result).encode(), "text/plain")
        )
        description = (
            self.description
            if isinstance(self.description, Unset)
            else (None, str(self.description).encode(), "text/plain")
        )
        output_data_type = (
            self.output_data_type
            if isinstance(self.output_data_type, Unset)
            else (None, str(self.output_data_type).encode(), "text/plain")
        )
        project_id = (
            self.project_id
            if isinstance(self.project_id, Unset)
            else (None, str(self.project_id).encode(), "text/plain")
        )
        file: Union[Unset, FileJsonType] = UNSET
        if not isinstance(self.file, Unset):
            file = self.file.to_tuple()

        update = self.update if isinstance(self.update, Unset) else (None, str(self.update).encode(), "text/plain")

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {key: (None, str(value).encode(), "text/plain") for key, value in self.additional_properties.items()}
        )
        field_dict.update(
            {
                "name": name,
                "requires_scenario_input": requires_scenario_input,
                "requires_scenario_result": requires_scenario_result,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if output_data_type is not UNSET:
            field_dict["output_data_type"] = output_data_type
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if file is not UNSET:
            field_dict["file"] = file
        if update is not UNSET:
            field_dict["update"] = update

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        name = d.pop("name")

        requires_scenario_input = d.pop("requires_scenario_input")

        requires_scenario_result = d.pop("requires_scenario_result")

        description = d.pop("description", UNSET)

        output_data_type = d.pop("output_data_type", UNSET)

        project_id = d.pop("project_id", UNSET)

        _file = d.pop("file", UNSET)
        file: Union[Unset, File]
        if isinstance(_file, Unset):
            file = UNSET
        else:
            file = File(payload=BytesIO(_file))

        update = d.pop("update", UNSET)

        body_check_upload_v0_check_upload_post = cls(
            name=name,
            requires_scenario_input=requires_scenario_input,
            requires_scenario_result=requires_scenario_result,
            description=description,
            output_data_type=output_data_type,
            project_id=project_id,
            file=file,
            update=update,
        )

        body_check_upload_v0_check_upload_post.additional_properties = d
        return body_check_upload_v0_check_upload_post

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ComparisonOperator(str, Enum):
    CONTAINS = "contains"
    EQUAL = "equal"
    GREATER_THAN = "greater_than"
    IS_SET = "is_set"
    LESS_THAN = "less_than"
    NOT_CONTAINS = "not_contains"
    NOT_EQUAL = "not_equal"
    NOT_SET = "not_set"

    def __str__(self) -> str:
        return str(self.value)



from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.error_response import ErrorResponse
from ...models.scenario_set_response import ScenarioSetResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    scenario_id: Union[Unset, None, str] = UNSET,
    project_id: Union[Unset, None, str] = UNSET,
    api_key: str,
) -> Dict[str, Any]:
    headers = {}
    headers["api-key"] = api_key

    params: Dict[str, Any] = {}
    params["scenario_id"] = scenario_id

    params["project_id"] = project_id

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    return {
        "method": "get",
        "url": "/v0/scenario_sets",
        "params": params,
        "headers": headers,
    }


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = ScenarioSetResponse.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ErrorResponse.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.UNPROCESSABLE_ENTITY:
        response_422 = ErrorResponse.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    scenario_id: Union[Unset, None, str] = UNSET,
    project_id: Union[Unset, None, str] = UNSET,
    api_key: str,
) -> Response[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    """Get Scenario Sets

     Find all scenario sets based on either the project id or the scenario id

    Returns:
        a list of scenario sets (project id) or a list of scenarios (scenario id)

    Args:
        scenario_id (Union[Unset, None, str]): The ID of the scenario set
        project_id (Union[Unset, None, str]): The ID of the project
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, List['ScenarioSetResponse']]]
    """

    kwargs = _get_kwargs(
        scenario_id=scenario_id,
        project_id=project_id,
        api_key=api_key,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    scenario_id: Union[Unset, None, str] = UNSET,
    project_id: Union[Unset, None, str] = UNSET,
    api_key: str,
) -> Optional[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    """Get Scenario Sets

     Find all scenario sets based on either the project id or the scenario id

    Returns:
        a list of scenario sets (project id) or a list of scenarios (scenario id)

    Args:
        scenario_id (Union[Unset, None, str]): The ID of the scenario set
        project_id (Union[Unset, None, str]): The ID of the project
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, List['ScenarioSetResponse']]
    """

    return sync_detailed(
        client=client,
        scenario_id=scenario_id,
        project_id=project_id,
        api_key=api_key,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    scenario_id: Union[Unset, None, str] = UNSET,
    project_id: Union[Unset, None, str] = UNSET,
    api_key: str,
) -> Response[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    """Get Scenario Sets

     Find all scenario sets based on either the project id or the scenario id

    Returns:
        a list of scenario sets (project id) or a list of scenarios (scenario id)

    Args:
        scenario_id (Union[Unset, None, str]): The ID of the scenario set
        project_id (Union[Unset, None, str]): The ID of the project
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[ErrorResponse, List['ScenarioSetResponse']]]
    """

    kwargs = _get_kwargs(
        scenario_id=scenario_id,
        project_id=project_id,
        api_key=api_key,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    scenario_id: Union[Unset, None, str] = UNSET,
    project_id: Union[Unset, None, str] = UNSET,
    api_key: str,
) -> Optional[Union[ErrorResponse, List["ScenarioSetResponse"]]]:
    """Get Scenario Sets

     Find all scenario sets based on either the project id or the scenario id

    Returns:
        a list of scenario sets (project id) or a list of scenarios (scenario id)

    Args:
        scenario_id (Union[Unset, None, str]): The ID of the scenario set
        project_id (Union[Unset, None, str]): The ID of the project
        api_key (str):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[ErrorResponse, List['ScenarioSetResponse']]
    """

    return (
        await asyncio_detailed(
            client=client,
            scenario_id=scenario_id,
            project_id=project_id,
            api_key=api_key,
        )
    ).parsed



import datetime
from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET, Unset

T = TypeVar("T", bound="DatapointSchema")


@_attrs_define
class DatapointSchema:
    """
    Attributes:
        mut_id (Union[Unset, str]): Model ID
        input_ (Union[Unset, str]): Inputted value into the model
        input_datetime (Union[Unset, datetime.datetime]): Datetime for the input
        result (Union[Unset, str]): Outputted value from the model based on the input
        result_datetime (Union[Unset, datetime.datetime]): Datetime for the result
        tags (Union[Unset, List[str]]): Tags are strings that can be used to filter datapoints in the Okareo app
        feedback (Union[Unset, float]): Feedback is a 0 to 1 float value that captures user feedback range for related
            datapoint results
        error_message (Union[Unset, str]):
        error_code (Union[Unset, str]):
        context_token (Union[Unset, str]): Context token is a unique token to link various datapoints which originate
            from the same context
        test_run_id (Union[Unset, str]): ID of testrun
        group_id (Union[Unset, str]): ID of the group
        model_metadata (Union[Unset, str]): Additional metadata about the model used for this datapoint
        input_metadata (Union[Unset, str]): Metadata about the input
        result_metadata (Union[Unset, str]): Metadata about the result
    """

    mut_id: Union[Unset, str] = UNSET
    input_: Union[Unset, str] = UNSET
    input_datetime: Union[Unset, datetime.datetime] = UNSET
    result: Union[Unset, str] = UNSET
    result_datetime: Union[Unset, datetime.datetime] = UNSET
    tags: Union[Unset, List[str]] = UNSET
    feedback: Union[Unset, float] = UNSET
    error_message: Union[Unset, str] = UNSET
    error_code: Union[Unset, str] = UNSET
    context_token: Union[Unset, str] = UNSET
    test_run_id: Union[Unset, str] = UNSET
    group_id: Union[Unset, str] = UNSET
    model_metadata: Union[Unset, str] = UNSET
    input_metadata: Union[Unset, str] = UNSET
    result_metadata: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        mut_id = self.mut_id
        input_ = self.input_
        input_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.input_datetime, Unset):
            input_datetime = self.input_datetime.isoformat()

        result = self.result
        result_datetime: Union[Unset, str] = UNSET
        if not isinstance(self.result_datetime, Unset):
            result_datetime = self.result_datetime.isoformat()

        tags: Union[Unset, List[str]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = self.tags

        feedback = self.feedback
        error_message = self.error_message
        error_code = self.error_code
        context_token = self.context_token
        test_run_id = self.test_run_id
        group_id = self.group_id
        model_metadata = self.model_metadata
        input_metadata = self.input_metadata
        result_metadata = self.result_metadata

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if mut_id is not UNSET:
            field_dict["mut_id"] = mut_id
        if input_ is not UNSET:
            field_dict["input"] = input_
        if input_datetime is not UNSET:
            field_dict["input_datetime"] = input_datetime
        if result is not UNSET:
            field_dict["result"] = result
        if result_datetime is not UNSET:
            field_dict["result_datetime"] = result_datetime
        if tags is not UNSET:
            field_dict["tags"] = tags
        if feedback is not UNSET:
            field_dict["feedback"] = feedback
        if error_message is not UNSET:
            field_dict["error_message"] = error_message
        if error_code is not UNSET:
            field_dict["error_code"] = error_code
        if context_token is not UNSET:
            field_dict["context_token"] = context_token
        if test_run_id is not UNSET:
            field_dict["test_run_id"] = test_run_id
        if group_id is not UNSET:
            field_dict["group_id"] = group_id
        if model_metadata is not UNSET:
            field_dict["model_metadata"] = model_metadata
        if input_metadata is not UNSET:
            field_dict["input_metadata"] = input_metadata
        if result_metadata is not UNSET:
            field_dict["result_metadata"] = result_metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        mut_id = d.pop("mut_id", UNSET)

        input_ = d.pop("input", UNSET)

        _input_datetime = d.pop("input_datetime", UNSET)
        input_datetime: Union[Unset, datetime.datetime]
        if isinstance(_input_datetime, Unset):
            input_datetime = UNSET
        else:
            input_datetime = isoparse(_input_datetime)

        result = d.pop("result", UNSET)

        _result_datetime = d.pop("result_datetime", UNSET)
        result_datetime: Union[Unset, datetime.datetime]
        if isinstance(_result_datetime, Unset):
            result_datetime = UNSET
        else:
            result_datetime = isoparse(_result_datetime)

        tags = cast(List[str], d.pop("tags", UNSET))

        feedback = d.pop("feedback", UNSET)

        error_message = d.pop("error_message", UNSET)

        error_code = d.pop("error_code", UNSET)

        context_token = d.pop("context_token", UNSET)

        test_run_id = d.pop("test_run_id", UNSET)

        group_id = d.pop("group_id", UNSET)

        model_metadata = d.pop("model_metadata", UNSET)

        input_metadata = d.pop("input_metadata", UNSET)

        result_metadata = d.pop("result_metadata", UNSET)

        datapoint_schema = cls(
            mut_id=mut_id,
            input_=input_,
            input_datetime=input_datetime,
            result=result,
            result_datetime=result_datetime,
            tags=tags,
            feedback=feedback,
            error_message=error_message,
            error_code=error_code,
            context_token=context_token,
            test_run_id=test_run_id,
            group_id=group_id,
            model_metadata=model_metadata,
            input_metadata=input_metadata,
            result_metadata=result_metadata,
        )

        datapoint_schema.additional_properties = d
        return datapoint_schema

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="TestRunPayloadV2MetricsKwargs")


@_attrs_define
class TestRunPayloadV2MetricsKwargs:
    """Dictionary of metrics to be measured"""

    additional_properties: Dict[str, Union[List[int], List[str]]] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        field_dict: Dict[str, Any] = {}
        for prop_name, prop in self.additional_properties.items():
            if isinstance(prop, list):
                field_dict[prop_name] = prop

            else:
                field_dict[prop_name] = prop

        field_dict.update({})

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        test_run_payload_v2_metrics_kwargs = cls()

        additional_properties = {}
        for prop_name, prop_dict in d.items():

            def _parse_additional_property(data: object) -> Union[List[int], List[str]]:
                try:
                    if not isinstance(data, list):
                        raise TypeError()
                    additional_property_type_0 = cast(List[int], data)

                    return additional_property_type_0
                except:  # noqa: E722
                    pass
                if not isinstance(data, list):
                    raise TypeError()
                additional_property_type_1 = cast(List[str], data)

                return additional_property_type_1

            additional_property = _parse_additional_property(prop_dict)

            additional_properties[prop_name] = additional_property

        test_run_payload_v2_metrics_kwargs.additional_properties = additional_properties
        return test_run_payload_v2_metrics_kwargs

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Union[List[int], List[str]]:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Union[List[int], List[str]]) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class DatapointField(str, Enum):
    CONTEXT_TOKEN = "context_token"
    COST = "cost"
    ERROR_CODE = "error_code"
    ERROR_MESSAGE = "error_message"
    ERROR_TYPE = "error_type"
    FEEDBACK = "feedback"
    ID = "id"
    INPUT = "input"
    INPUT_DATETIME = "input_datetime"
    INPUT_TOKENS = "input_tokens"
    INPUT_TOOLS = "input_tools"
    ISSUE_TYPE = "issue_type"
    LATENCY = "latency"
    MODEL_METADATA = "model_metadata"
    OUTPUT_TOKENS = "output_tokens"
    PROVIDER = "provider"
    REQUEST_MODEL_NAME = "request_model_name"
    RESPONSE_MODEL_NAME = "response_model_name"
    RESULT = "result"
    RESULT_DATETIME = "result_datetime"
    RESULT_TOOL_CALLS = "result_tool_calls"
    SOURCE = "source"
    TAGS = "tags"
    TEMPERATURE = "temperature"
    TEST_DATA_POINT_ID = "test_data_point_id"
    TEST_RUN_ID = "test_run_id"
    TIME_CREATED = "time_created"
    USER_METADATA = "user_metadata"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.generation_tone import GenerationTone
from ..models.scenario_type import ScenarioType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.scenario_data_poin_response import ScenarioDataPoinResponse
    from ..models.scenario_set_generate_checks_item_type_1 import ScenarioSetGenerateChecksItemType1
    from ..models.scenario_set_generate_generation_schema import ScenarioSetGenerateGenerationSchema


T = TypeVar("T", bound="ScenarioSetGenerate")


@_attrs_define
class ScenarioSetGenerate:
    """
    Attributes:
        name (str): Name of the generated scenario set
        number_examples (int): Number of examples to be generated for the scenario set
        project_id (Union[Unset, str]): ID for the project
        source_scenario_id (Union[Unset, str]): ID for the scenario set that the generated scenario set will use as a
            source. Will throw an exception if 'source_scenario_rows' is also provided.
        source_scenario_rows (Union[Unset, List['ScenarioDataPoinResponse']]): Rows for the scenario set that the
            generated scenario set will use as a source. Will throw an exception if 'source_scenario_id' is also provided.
        synonym_sets (Union[Unset, List[List[str]]]): 2D list used by the generator to determine synonyms. Used with the
            SYNONYMS generation type.
        save_generated_scenario (Union[Unset, bool]): Whether to save the generated scenarios. Defaults to True.
            Default: True.
        generation_type (Union[Unset, ScenarioType]): An enumeration. Default: ScenarioType.REPHRASE_INVARIANT.
        generation_tone (Union[Unset, GenerationTone]): An enumeration. Default: GenerationTone.NEUTRAL.
        generation_prompt (Union[Unset, str]): Prompt for the generator to use when generating scenarios. Only supported
            by CustomGenerator type.
        generation_schema (Union[Unset, ScenarioSetGenerateGenerationSchema]): Structured output schema for the
            generator to use when generating scenarios. Only supported by CustomGenerator type.
        pre_template (Union[Unset, str]): Template for pre-processing scenario before sending it to generator
        post_template (Union[Unset, str]): Template for post-processing scenario after generator before it's saved
        lock_result (Union[Unset, bool]): Whether to lock the result of the generated scenario. Used in the Custom
            Generator type.
        checks (Union[Unset, List[Union['ScenarioSetGenerateChecksItemType1', str]]]): List of check names or check
            configs to run on the generated scenarios
    """

    name: str
    number_examples: int
    project_id: Union[Unset, str] = UNSET
    source_scenario_id: Union[Unset, str] = UNSET
    source_scenario_rows: Union[Unset, List["ScenarioDataPoinResponse"]] = UNSET
    synonym_sets: Union[Unset, List[List[str]]] = UNSET
    save_generated_scenario: Union[Unset, bool] = True
    generation_type: Union[Unset, ScenarioType] = ScenarioType.REPHRASE_INVARIANT
    generation_tone: Union[Unset, GenerationTone] = GenerationTone.NEUTRAL
    generation_prompt: Union[Unset, str] = UNSET
    generation_schema: Union[Unset, "ScenarioSetGenerateGenerationSchema"] = UNSET
    pre_template: Union[Unset, str] = UNSET
    post_template: Union[Unset, str] = UNSET
    lock_result: Union[Unset, bool] = False
    checks: Union[Unset, List[Union["ScenarioSetGenerateChecksItemType1", str]]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.scenario_set_generate_checks_item_type_1 import ScenarioSetGenerateChecksItemType1

        name = self.name
        number_examples = self.number_examples
        project_id = self.project_id
        source_scenario_id = self.source_scenario_id
        source_scenario_rows: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.source_scenario_rows, Unset):
            source_scenario_rows = []
            for source_scenario_rows_item_data in self.source_scenario_rows:
                source_scenario_rows_item = source_scenario_rows_item_data.to_dict()

                source_scenario_rows.append(source_scenario_rows_item)

        synonym_sets: Union[Unset, List[List[str]]] = UNSET
        if not isinstance(self.synonym_sets, Unset):
            synonym_sets = []
            for synonym_sets_item_data in self.synonym_sets:
                synonym_sets_item = synonym_sets_item_data

                synonym_sets.append(synonym_sets_item)

        save_generated_scenario = self.save_generated_scenario
        generation_type: Union[Unset, str] = UNSET
        if not isinstance(self.generation_type, Unset):
            generation_type = self.generation_type.value

        generation_tone: Union[Unset, str] = UNSET
        if not isinstance(self.generation_tone, Unset):
            generation_tone = self.generation_tone.value

        generation_prompt = self.generation_prompt
        generation_schema: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.generation_schema, Unset):
            generation_schema = self.generation_schema.to_dict()

        pre_template = self.pre_template
        post_template = self.post_template
        lock_result = self.lock_result
        checks: Union[Unset, List[Union[Dict[str, Any], str]]] = UNSET
        if not isinstance(self.checks, Unset):
            checks = []
            for checks_item_data in self.checks:
                checks_item: Union[Dict[str, Any], str]

                if isinstance(checks_item_data, ScenarioSetGenerateChecksItemType1):
                    checks_item = checks_item_data.to_dict()

                else:
                    checks_item = checks_item_data

                checks.append(checks_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "number_examples": number_examples,
            }
        )
        if project_id is not UNSET:
            field_dict["project_id"] = project_id
        if source_scenario_id is not UNSET:
            field_dict["source_scenario_id"] = source_scenario_id
        if source_scenario_rows is not UNSET:
            field_dict["source_scenario_rows"] = source_scenario_rows
        if synonym_sets is not UNSET:
            field_dict["synonym_sets"] = synonym_sets
        if save_generated_scenario is not UNSET:
            field_dict["save_generated_scenario"] = save_generated_scenario
        if generation_type is not UNSET:
            field_dict["generation_type"] = generation_type
        if generation_tone is not UNSET:
            field_dict["generation_tone"] = generation_tone
        if generation_prompt is not UNSET:
            field_dict["generation_prompt"] = generation_prompt
        if generation_schema is not UNSET:
            field_dict["generation_schema"] = generation_schema
        if pre_template is not UNSET:
            field_dict["pre_template"] = pre_template
        if post_template is not UNSET:
            field_dict["post_template"] = post_template
        if lock_result is not UNSET:
            field_dict["lock_result"] = lock_result
        if checks is not UNSET:
            field_dict["checks"] = checks

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.scenario_data_poin_response import ScenarioDataPoinResponse
        from ..models.scenario_set_generate_checks_item_type_1 import ScenarioSetGenerateChecksItemType1
        from ..models.scenario_set_generate_generation_schema import ScenarioSetGenerateGenerationSchema

        d = src_dict.copy()
        name = d.pop("name")

        number_examples = d.pop("number_examples")

        project_id = d.pop("project_id", UNSET)

        source_scenario_id = d.pop("source_scenario_id", UNSET)

        source_scenario_rows = []
        _source_scenario_rows = d.pop("source_scenario_rows", UNSET)
        for source_scenario_rows_item_data in _source_scenario_rows or []:
            source_scenario_rows_item = ScenarioDataPoinResponse.from_dict(source_scenario_rows_item_data)

            source_scenario_rows.append(source_scenario_rows_item)

        synonym_sets = []
        _synonym_sets = d.pop("synonym_sets", UNSET)
        for synonym_sets_item_data in _synonym_sets or []:
            synonym_sets_item = cast(List[str], synonym_sets_item_data)

            synonym_sets.append(synonym_sets_item)

        save_generated_scenario = d.pop("save_generated_scenario", UNSET)

        _generation_type = d.pop("generation_type", UNSET)
        generation_type: Union[Unset, ScenarioType]
        if isinstance(_generation_type, Unset):
            generation_type = UNSET
        else:
            generation_type = ScenarioType(_generation_type)

        _generation_tone = d.pop("generation_tone", UNSET)
        generation_tone: Union[Unset, GenerationTone]
        if isinstance(_generation_tone, Unset):
            generation_tone = UNSET
        else:
            generation_tone = GenerationTone(_generation_tone)

        generation_prompt = d.pop("generation_prompt", UNSET)

        _generation_schema = d.pop("generation_schema", UNSET)
        generation_schema: Union[Unset, ScenarioSetGenerateGenerationSchema]
        if isinstance(_generation_schema, Unset):
            generation_schema = UNSET
        else:
            generation_schema = ScenarioSetGenerateGenerationSchema.from_dict(_generation_schema)

        pre_template = d.pop("pre_template", UNSET)

        post_template = d.pop("post_template", UNSET)

        lock_result = d.pop("lock_result", UNSET)

        checks = []
        _checks = d.pop("checks", UNSET)
        for checks_item_data in _checks or []:

            def _parse_checks_item(data: object) -> Union["ScenarioSetGenerateChecksItemType1", str]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    checks_item_type_1 = ScenarioSetGenerateChecksItemType1.from_dict(data)

                    return checks_item_type_1
                except:  # noqa: E722
                    pass
                return cast(Union["ScenarioSetGenerateChecksItemType1", str], data)

            checks_item = _parse_checks_item(checks_item_data)

            checks.append(checks_item)

        scenario_set_generate = cls(
            name=name,
            number_examples=number_examples,
            project_id=project_id,
            source_scenario_id=source_scenario_id,
            source_scenario_rows=source_scenario_rows,
            synonym_sets=synonym_sets,
            save_generated_scenario=save_generated_scenario,
            generation_type=generation_type,
            generation_tone=generation_tone,
            generation_prompt=generation_prompt,
            generation_schema=generation_schema,
            pre_template=pre_template,
            post_template=post_template,
            lock_result=lock_result,
            checks=checks,
        )

        scenario_set_generate.additional_properties = d
        return scenario_set_generate

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
import os

from okareo_api_client.models import TestRunItem


class JSONReporter:
    """
    A simple class to define a json reporter for output of results.
    This class assumes that it is being used within the context of the CLI.
    """

    def __init__(self, eval_runs: list[TestRunItem]):
        self.eval_runs = eval_runs

    def log(self) -> None:
        """
        Logs the results of the evaluation runs into a file based on OKAREO_REPORT_DIR.
        """
        if len(self.eval_runs) == 0:
            raise ValueError("No evaluation runs to log.")

        okareo_report_dir = os.getenv("OKAREO_REPORT_DIR")
        if (
            okareo_report_dir is not None
            and len(okareo_report_dir) > 0
            and not os.path.exists(okareo_report_dir)
        ):
            os.makedirs(okareo_report_dir, exist_ok=True)

        try:
            for eval_item in self.eval_runs:
                eval_dict = eval_item.to_dict()
                if okareo_report_dir is None or len(okareo_report_dir) == 0:
                    print(json.dumps(eval_dict, indent=2))
                    continue

                eval_file_name: str = eval_item.id
                if eval_item.name not in (None, ""):
                    eval_file_name = str(eval_item.name)
                file_name = eval_file_name.replace(" ", "_")
                file_path = os.path.join(okareo_report_dir, f"{file_name}.json")
                with open(file_path, "w") as outfile:
                    json.dump(eval_dict, outfile, indent=2)

        except Exception as e:
            print(f"Unexpected result {e=}")
            raise

        return



import os
from datetime import datetime

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.model_under_test import GenerationModel
from okareo_api_client.models import ScenarioSetResponse, TestRunType
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.test_run_item_model_metrics import TestRunItemModelMetrics

today_with_time = datetime.now().strftime("%m-%d %H:%M:%S")
rnd_str = random_string(5)
unique_key = f"{rnd_str} {today_with_time}"
create_scenario_name = f"ci_test_dialog_scenarios {unique_key}"


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


def validate_scenario_set(
    scenario_set_validated: ScenarioSetResponse, name: str, okareo_client: Okareo
) -> None:
    assert scenario_set_validated is not None
    assert scenario_set_validated.name == name
    assert scenario_set_validated.scenario_id
    assert scenario_set_validated.project_id
    assert scenario_set_validated.time_created

    scenario_dps = okareo_client.get_scenario_data_points(
        scenario_set_validated.scenario_id
    )
    assert scenario_dps
    assert len(scenario_dps) == 1
    for dp in scenario_dps:
        input_ = dp.to_dict()["input"]
        assert input_
        assert input_[0]["role"] == "system"
        assert input_[1]["role"] == "user"
        assert input_[2]["role"] == "assistant"
        assert input_[3]["role"] == "function"


JSON_DAILOG = Okareo.seed_data_from_list(
    [
        {
            "input": [
                {"role": "system", "content": "You are super genious mind reader!"},
                {
                    "role": "user",
                    "content": (
                        "I want to know the temperature in location i'm thinking of ... "
                        + "figure out the location and temperature!"
                    ),
                },
                {"role": "assistant", "content": "I can help you with that!"},
                {
                    "role": "function",
                    "name": "get_temp",
                    "content": "Error from server (NotFound)",
                },
            ],
            "result": {"role": "assistant", "content": "Oops!"},
        },
    ]
)


@pytest.fixture(scope="module")
def create_scenario_set(okareo_client: Okareo) -> ScenarioSetResponse:
    scenario_set_create = ScenarioSetCreate(
        name=create_scenario_name,
        seed_data=JSON_DAILOG,
    )
    scenario = okareo_client.create_scenario_set(scenario_set_create)
    return scenario


def test_create_scenario_set(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> None:
    validate_scenario_set(create_scenario_set, create_scenario_name, okareo_client)


def test_dialog_input(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> None:
    test_run_name = f"ci_test_dialog_scenarios_openai {unique_key}"

    mut = okareo_client.register_model(
        name=test_run_name,
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            dialog_template="{scenario_input}",
        ),
    )

    run_resp = mut.run_test(
        name=test_run_name,
        scenario=create_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=["compression_ratio", "levenshtein_distance"],
    )
    assert run_resp.name == test_run_name
    assert isinstance(run_resp.model_metrics, TestRunItemModelMetrics)
    metrics_dict = run_resp.model_metrics.to_dict()
    assert metrics_dict["mean_scores"] is not None
    assert metrics_dict["scores_by_row"] is not None


DAILOG_TEMPLATE = """
[
    {
        "role": "{scenario_input.0.role}",
        "content": "{scenario_input.0.content}"
    },
    {
        "role": "{scenario_input.2.role}",
        "content": "{scenario_input.2.content}"
    }
]
"""


def test_dialog_template(
    okareo_client: Okareo, create_scenario_set: ScenarioSetResponse
) -> None:
    test_run_name = f"ci_test_dialog_template_openai {unique_key}"

    mut = okareo_client.register_model(
        name=test_run_name,
        model=GenerationModel(
            model_id="gpt-4.1-mini",
            temperature=0,
            dialog_template=DAILOG_TEMPLATE,
        ),
    )

    run_resp = mut.run_test(
        name=test_run_name,
        scenario=create_scenario_set,
        api_key=os.environ["OPENAI_API_KEY"],
        test_run_type=TestRunType.NL_GENERATION,
        checks=["compression_ratio", "levenshtein_distance"],
    )
    assert run_resp.name == test_run_name
    assert isinstance(run_resp.model_metrics, TestRunItemModelMetrics)
    metrics_dict = run_resp.model_metrics.to_dict()
    assert metrics_dict["mean_scores"] is not None
    assert metrics_dict["scores_by_row"] is not None



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="UpdateTestDataPointPayload")


@_attrs_define
class UpdateTestDataPointPayload:
    """
    Attributes:
        ids (Union[Unset, List[str]]): IDs of the datapoints to update
        tags (Union[Unset, List[List[str]]]): Tags are strings that can be used to filter test data points in the Okareo
            app
    """

    ids: Union[Unset, List[str]] = UNSET
    tags: Union[Unset, List[List[str]]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.ids, Unset):
            ids = self.ids

        tags: Union[Unset, List[List[str]]] = UNSET
        if not isinstance(self.tags, Unset):
            tags = []
            for tags_item_data in self.tags:
                tags_item = tags_item_data

                tags.append(tags_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if ids is not UNSET:
            field_dict["ids"] = ids
        if tags is not UNSET:
            field_dict["tags"] = tags

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        ids = cast(List[str], d.pop("ids", UNSET))

        tags = []
        _tags = d.pop("tags", UNSET)
        for tags_item_data in _tags or []:
            tags_item = cast(List[str], tags_item_data)

            tags.append(tags_item)

        update_test_data_point_payload = cls(
            ids=ids,
            tags=tags,
        )

        update_test_data_point_payload.additional_properties = d
        return update_test_data_point_payload

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os
from datetime import datetime
from typing import Optional
from unittest.mock import Mock

import pytest
from okareo_tests.common import API_KEY, OkareoAPIhost, integration, random_string
from pytest_httpx import HTTPXMock

from okareo import ModelUnderTest, Okareo
from okareo.error import TestRunError
from okareo.model_under_test import CohereModel, OpenAIModel, PineconeDb
from okareo_api_client.models import SeedData
from okareo_api_client.models.scenario_set_create import ScenarioSetCreate
from okareo_api_client.models.test_run_type import TestRunType

GLOBAL_PROJECT_RESPONSE = [
    {
        "id": "0156f5d7-4ac4-4568-9d44-24750aa08d1a",
        "name": "Global",
        "onboarding_status": "onboarding_status",
        "tags": [],
        "additional_properties": {},
    }
]


def helper_register_model(httpx_mock: HTTPXMock) -> ModelUnderTest:
    httpx_mock.add_response(
        json=GLOBAL_PROJECT_RESPONSE,
        status_code=201,
    )
    fixture = get_mut_fixture("NotebookModel")
    httpx_mock.add_response(status_code=201, json=fixture)
    okareo = Okareo("api-key", "http://mocked.com")
    return okareo.register_model(name="NotebookModel", tags=["ci-testing"])


def get_mut_fixture(name: Optional[str] = None) -> dict:
    rnd_str = random_string(5)
    return {
        "id": "1",
        "project_id": "1",
        "name": name if name else f"CI-Test-Model-{rnd_str}",
        "tags": ["ci-testing"],
        "time_created": "foo",
    }


@integration
def test_register_model(httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost) -> None:
    fixture = get_mut_fixture()
    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=GLOBAL_PROJECT_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(status_code=201, json=fixture)

    okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
    mut = okareo.register_model(name=fixture["name"], tags=fixture["tags"])
    assert mut.name == fixture["name"]


@integration
def test_add_datapoint_optional_integration(
    httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost
) -> None:
    fixture = get_mut_fixture()
    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=GLOBAL_PROJECT_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(status_code=201, json=fixture)

    okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
    mut = okareo.register_model(name=fixture["name"], tags=fixture["tags"])
    dp = mut.add_data_point(
        feedback=0,
        context_token="SOME_CONTEXT_TOKEN",
    )
    assert dp


@integration
def test_mut_test_run(httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost) -> None:
    mut_fixture = get_mut_fixture()

    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=GLOBAL_PROJECT_RESPONSE,
            status_code=201,
        )
        httpx_mock.add_response(
            status_code=201,
            json={
                "scenario_id": "scenario-id",
                "project_id": "project-id",
                "time_created": str(datetime.now()),
                "type": "REPHRASE_INVARIANT",
            },
        )
        httpx_mock.add_response(status_code=201, json=mut_fixture)
        httpx_mock.add_response(
            status_code=201,
            json={
                "id": "test-run-item-id",
                "name": "CI run test",
                "input": "",
                "result": "",
                "project_id": "",
                "mut_id": "",
                "scenario_set_id": "",
            },
        )

    okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)

    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=f"my test scenario set {random_string(10)}",
        seed_data=[
            SeedData(input_="example question or statement", result="example result")
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)
    response.scenario_id

    mut = okareo.register_model(
        name=mut_fixture["name"],
        tags=mut_fixture["tags"],
        model=OpenAIModel(
            model_id="gpt-4o-mini",
            temperature=1,
            system_prompt_template="system_prompt_template",
        ),
    )

    # use the scenario id from one of the scenario set notebook examples
    test_run_item = mut.run_test(
        scenario=response, name="CI run test", api_key=os.environ["OPENAI_API_KEY"]
    )
    assert test_run_item.name == "CI run test"


@integration
def test_mut_test_run_with_id(httpx_mock: HTTPXMock, okareo_api: OkareoAPIhost) -> None:
    mut_fixture = get_mut_fixture()

    if okareo_api.is_mock:
        httpx_mock.add_response(
            json=GLOBAL_PROJECT_RESPONSE,
            status_code=201,
        )

        httpx_mock.add_response(
            status_code=201,
            json={
                "scenario_id": "scenario-id",
                "project_id": "project-id",
                "time_created": str(datetime.now()),
                "type": "REPHRASE_INVARIANT",
            },
        )

        httpx_mock.add_response(status_code=201, json=mut_fixture)
        httpx_mock.add_response(
            status_code=201,
            json={
                "id": "test-run-item-id",
                "name": "CI run test",
                "input": "",
                "result": "",
                "project_id": "",
                "mut_id": "",
                "scenario_set_id": "",
            },
        )

    okareo = Okareo(api_key=API_KEY, base_path=okareo_api.path)
    # generate scenario and return results in one call
    scenario_set_create = ScenarioSetCreate(
        name=f"my test scenario set {random_string(10)}",
        seed_data=[
            SeedData(input_="example question or statement", result="example result")
        ],
    )
    response = okareo.create_scenario_set(scenario_set_create)

    mut = okareo.register_model(
        name=mut_fixture["name"],
        tags=mut_fixture["tags"],
        model=OpenAIModel(
            model_id="gpt-4o-mini",
            temperature=1,
            system_prompt_template="system_prompt_template",
        ),
    )

    # use the scenario id from one of the scenario set notebook examples
    assert isinstance(response.scenario_id, str)
    test_run_item = mut.run_test(
        scenario=response.scenario_id,
        name="CI run test",
        api_key=os.environ["OPENAI_API_KEY"],
    )
    assert test_run_item.name == "CI run test"


def test_add_datapoint(httpx_mock: HTTPXMock) -> None:
    registered_model = helper_register_model(httpx_mock)
    fixture = {"id": "1", "project_id": "1", "mut_id": "1"}

    httpx_mock.add_response(status_code=201, json=fixture)
    dp = registered_model.add_data_point(
        input_obj={"input": "value"},
        result_obj={"result": "value"},
        feedback=0,
        context_token="SOME_CONTEXT_TOKEN",
        input_datetime=str(datetime.now()),
        result_datetime=str(datetime.now()),
        tags=["ci-testing"],
    )
    assert dp


def test_add_datapoint_only_context_and_feedback(httpx_mock: HTTPXMock) -> None:
    registered_model = helper_register_model(httpx_mock)
    fixture = {"id": "1", "project_id": "1", "mut_id": "1"}
    httpx_mock.add_response(status_code=201, json=fixture)
    dp = registered_model.add_data_point(
        feedback=0,
        context_token="SOME_CONTEXT_TOKEN",
    )
    assert dp


def test_get_test_run(httpx_mock: HTTPXMock) -> None:
    registered_model = helper_register_model(httpx_mock)

    # Mocking response for the get_test_run_v0_test_runs_test_run_id_get API call
    httpx_mock.add_response(
        status_code=201,
        json={
            "id": "test_id",
            "project_id": "project_id",
            "mut_id": "mut_id",
            "scenario_set_id": "scenario_set_id",
        },
    )

    result = registered_model.get_test_run("test_run_id")
    assert result


def test_missing_api_key_test_run_modelv2(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        json=GLOBAL_PROJECT_RESPONSE,
        status_code=201,
    )
    httpx_mock.add_response(status_code=201, json=get_mut_fixture())
    okareo = Okareo("api-key", "http://mocked.com")
    mut = okareo.register_model(
        name="complex model test",
        model=[
            CohereModel(model_id="foo", model_type="embed"),
            PineconeDb(index_name="bar", region="baz", project_id="fooz"),
        ],
    )

    with pytest.raises(TestRunError):
        mut.run_test(
            name="quick test",
            scenario=Mock(),
            calculate_metrics=True,
            test_run_type=TestRunType.INFORMATION_RETRIEVAL,
            api_keys={"cohere": "foo"},
        )


def test_missing_vector_db_key_test_run_modelv2(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(
        json=GLOBAL_PROJECT_RESPONSE,
        status_code=201,
    )
    httpx_mock.add_response(status_code=201, json=get_mut_fixture())
    okareo = Okareo("api-key", "http://mocked.com")
    mut = okareo.register_model(
        name="complex model test",
        model=[
            CohereModel(model_id="foo", model_type="embed"),
        ],
    )

    with pytest.raises(TestRunError):
        mut.run_test(
            name="quick test",
            scenario=Mock(),
            calculate_metrics=True,
            test_run_type=TestRunType.INFORMATION_RETRIEVAL,
            api_keys={"cohere": "foo"},
        )



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="FeedbackRangeSummary")


@_attrs_define
class FeedbackRangeSummary:
    """
    Attributes:
        count (int): The total count of feedbacks in the specified range for the given date.
        feedback_range_start (Union[Unset, float]): The start of the feedback range.
    """

    count: int
    feedback_range_start: Union[Unset, float] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        count = self.count
        feedback_range_start = self.feedback_range_start

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "count": count,
            }
        )
        if feedback_range_start is not UNSET:
            field_dict["feedback_range_start"] = feedback_range_start

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        count = d.pop("count")

        feedback_range_start = d.pop("feedback_range_start", UNSET)

        feedback_range_summary = cls(
            count=count,
            feedback_range_start=feedback_range_start,
        )

        feedback_range_summary.additional_properties = d
        return feedback_range_summary

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
import importlib.metadata
import json
import logging
import random
import string
import uuid
from typing import Any, Collection, Optional

from opentelemetry import trace
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor  # type: ignore
from opentelemetry.sdk.trace import ReadableSpan, SpanProcessor, TracerProvider
from opentelemetry.trace import get_tracer
from wrapt import wrap_function_wrapper  # type: ignore

from okareo import Okareo

logging.basicConfig(level=logging.FATAL)


def get_agent_names(json_data: Any) -> Any:
    data = json.loads(json_data)
    agents = data.get("crew_agents", [])
    agent_names = [agent["role"] for agent in agents]
    return agent_names


def find_agent_role(json_data: Any, id_prefix: Any) -> Any:
    data = json.loads(json_data)

    for task in data["crew_tasks"]:
        if task["id"].startswith(id_prefix):
            role = task["agent_role"]
            return role
    return None


class CrewAISpanProcessor(SpanProcessor):
    def __init__(self, config: dict[str, Any]) -> None:
        if "api_key" not in config:
            raise ValueError("api_key is required in the config")
        base_path = config.get("base_path", None)
        api_key = config["api_key"]

        if base_path and len(base_path) > 0:
            self.okareo = Okareo(api_key, base_path=base_path)
        else:
            self.okareo = Okareo(api_key)
        self.is_context_set = "context_token" in config
        self.context_id = str(config.get("context_token", uuid.uuid4()))
        self.tags = config.get("tags", [])
        random_suffix = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=5)
        )
        self.group_name = config.get("group_name", f"crewai-chat-{random_suffix}")

        self.group = self.okareo.create_group(
            name=self.group_name, tags=self.tags, source={"log_source": "crewai"}
        )
        self.created_obj: Any = {}
        self.cur_model: Any = None
        self.total_task_count = 0
        self.task_count = 0

    def _format_span_data(self, span: ReadableSpan) -> dict:
        if span.attributes is None:
            return {}
        formatted_data = {
            k: self.safe_loads(v)
            for k, v in span.attributes.items()
            if self.is_serializable(self.safe_loads(v))
        }
        if "openai.message_get" in formatted_data:
            formatted_data["openai.message_get"] = self.safe_loads(
                formatted_data["openai.message_get"]
            )
        return formatted_data

    def _get_timestamp_iso(self, epoch_ns: Any) -> str:
        timestamp = datetime.datetime.fromtimestamp(
            epoch_ns / 1e9, tz=datetime.timezone.utc
        ).isoformat()
        return timestamp

    def on_end(self, span: ReadableSpan) -> None:
        if span.name == "Crew Created":
            self.task_count = 0
            self.total_task_count = 0
            if not self.is_context_set:
                self.context_id = str(uuid.uuid4())
            self.created_obj = self._format_span_data(span)
            self.total_task_count = len(self.created_obj["crew_tasks"])

            agent_names = get_agent_names(json.dumps(self.created_obj))
            self.models = []
            for agent_name in agent_names:
                model = self.okareo.register_model(name=agent_name)
                self.models.append(model)
                self.okareo.add_model_to_group(self.group, model)
        models_to_add = []
        span_data = self._format_span_data(span)
        if span.name == "Crew Created":
            models_to_add = self.models
        elif span.name == "Task Created":
            role = find_agent_role(
                json.dumps(self.created_obj), span_data.get("task_id")
            )
            self.cur_model = self.okareo.register_model(name=role)
            models_to_add = [self.cur_model]
        else:
            models_to_add = [self.cur_model]

        result_obj = {
            "log_type": span.name,
            "trace_id": format(span.context.trace_id, "032x"),
            "span_id": format(span.context.span_id, "016x"),
        }
        for cur_model in models_to_add:
            if cur_model:
                cur_model.add_data_point(
                    input_obj=span_data,
                    input_datetime=self._get_timestamp_iso(span.start_time),
                    result_obj=json.dumps(result_obj),
                    result_datetime=self._get_timestamp_iso(span.end_time),
                    context_token=self.context_id,
                    tags=self.tags,
                    group_id=self.group.get("id"),
                )
        if span.name == "Task Execution":
            self.task_count += 1
            if self.task_count == self.total_task_count:
                self.okareo.create_trace_eval(self.group, self.context_id)

    @staticmethod
    def safe_loads(data: Any) -> Any:
        try:
            loaded_data = json.loads(data)
            return loaded_data
        except (json.JSONDecodeError, TypeError):
            return data

    @staticmethod
    def is_serializable(obj: Any) -> bool:
        if isinstance(obj, (str, int, float, bool, type(None))):
            return True
        if isinstance(obj, dict):
            return all(CrewAISpanProcessor.is_serializable(v) for v in obj.values())
        if isinstance(obj, (list, tuple)):
            return all(CrewAISpanProcessor.is_serializable(item) for item in obj)
        return False


def completion_wrapper(version: Any, tracer: Any) -> Any:
    def wrapper(wrapped: Any, instance: Any, args: Any, kwargs: Any) -> Any:
        with tracer.start_as_current_span("litellm.completion") as span:
            span.set_attribute("litellm.version", version)

            model = kwargs.get("model") or args[0] if args else None
            if model:
                span.set_attribute("litellm.model", model)

            messages = kwargs.get("messages")
            if messages:
                messages_json = json.dumps(messages, indent=2)
                span.set_attribute("messages", messages_json)

            try:
                result = wrapped(*args, **kwargs)

                if isinstance(result, dict):
                    usage = result.get("usage", {})
                    span.set_attribute(
                        "litellm.completion_tokens",
                        usage.get("completion_tokens"),
                    )
                    span.set_attribute(
                        "litellm.prompt_tokens",
                        usage.get("prompt_tokens"),
                    )
                    span.set_attribute(
                        "litellm.total_tokens",
                        usage.get("total_tokens"),
                    )

                return result
            except Exception as e:
                span.record_exception(e)

                raise

    return wrapper


class LiteLLMInstrumentation(BaseInstrumentor):  # type: ignore
    def instrumentation_dependencies(self) -> Collection[str]:
        return ["litellm >= 1.48.0", "opentelemetry-api >= 1.0.0"]

    def _instrument(self, **kwargs: Any) -> None:
        tracer_provider: Optional[TracerProvider] = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "", tracer_provider)
        version: str = importlib.metadata.version("litellm")
        wrap_function_wrapper(
            "litellm",
            "completion",
            completion_wrapper(version, tracer),
        )

        wrap_function_wrapper(
            "litellm",
            "text_completion",
            completion_wrapper(version, tracer),
        )

    def _uninstrument(self, **kwargs: Any) -> None:
        pass


def chat_completions_create(version: Any, tracer: Any) -> Any:
    def wrapper(wrapped: Any, instance: Any, args: Any, kwargs: Any) -> Any:
        with tracer.start_as_current_span("openai.chat.completions.create") as span:
            span.set_attribute("openai.version", version)

            model = kwargs.get("model")
            if model:
                span.set_attribute("openai.model", model)

            messages = kwargs.get("messages")
            span.set_attribute("openai.message_get", json.dumps(messages))

            try:
                result = wrapped(*args, **kwargs)

                if hasattr(result, "usage"):
                    span.set_attribute(
                        "openai.completion_tokens", result.usage.completion_tokens
                    )
                    span.set_attribute(
                        "openai.prompt_tokens", result.usage.prompt_tokens
                    )
                    span.set_attribute("openai.total_tokens", result.usage.total_tokens)

                return result
            except Exception as e:
                span.record_exception(e)
                raise

    return wrapper


class OpenAIInstrumentation(BaseInstrumentor):  # type: ignore
    def instrumentation_dependencies(self) -> Collection[str]:
        return ["openai >= 0.27.0", "opentelemetry-api >= 1.0.0"]

    def _instrument(self, **kwargs: Any) -> None:
        tracer_provider: Optional[TracerProvider] = kwargs.get("tracer_provider")
        tracer = get_tracer(__name__, "", tracer_provider)
        version: str = importlib.metadata.version("openai")

        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "Completions.create",
            chat_completions_create(version, tracer),
        )
        wrap_function_wrapper(
            "openai.resources.chat.completions",
            "AsyncCompletions.create",
            chat_completions_create(version, tracer),
        )

    def _uninstrument(self, **kwargs: Any) -> None:
        pass


class CrewAILogger:
    _instance = None
    _provider = None

    def __new__(cls, *args, **kwargs) -> Any:  # type: ignore
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, logger_config: dict[str, Any]) -> None:
        self.logger_config: Any = logger_config
        self.span_processor: Any = None

    def start(self) -> None:
        if CrewAILogger._provider is None:
            CrewAILogger._provider = TracerProvider()
            trace.set_tracer_provider(CrewAILogger._provider)
            LiteLLMInstrumentation().instrument(
                tracer_provider=CrewAILogger._provider, skip_dep_check=True
            )
            OpenAIInstrumentation().instrument(
                tracer_provider=CrewAILogger._provider, skip_dep_check=True
            )

        if self.span_processor is None:
            self.span_processor = CrewAISpanProcessor(self.logger_config)
            CrewAILogger._provider.add_span_processor(self.span_processor)

    def stop(self) -> None:
        if self.span_processor:
            self.span_processor.shutdown()
            self.span_processor = None
        # Note: We're not removing the span processor from the TracerProvider
        # as there's no built-in method to do so.

    def __enter__(self) -> "CrewAILogger":
        self.start()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.stop()

    def __del__(self) -> None:
        self.stop()



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.model_under_test_response_models import ModelUnderTestResponseModels


T = TypeVar("T", bound="ModelUnderTestResponse")


@_attrs_define
class ModelUnderTestResponse:
    """
    Attributes:
        id (str):
        project_id (str):
        name (str):
        tags (List[str]):
        time_created (str):
        models (Union[Unset, ModelUnderTestResponseModels]):
        sensitive_fields (Union[Unset, List[str]]):
        datapoint_count (Union[Unset, int]):
        app_link (Union[Unset, str]): This URL links to the Okareo webpage for this model Default: ''.
        warning (Union[Unset, str]):
    """

    id: str
    project_id: str
    name: str
    tags: List[str]
    time_created: str
    models: Union[Unset, "ModelUnderTestResponseModels"] = UNSET
    sensitive_fields: Union[Unset, List[str]] = UNSET
    datapoint_count: Union[Unset, int] = UNSET
    app_link: Union[Unset, str] = ""
    warning: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        id = self.id
        project_id = self.project_id
        name = self.name
        tags = self.tags

        time_created = self.time_created
        models: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.models, Unset):
            models = self.models.to_dict()

        sensitive_fields: Union[Unset, List[str]] = UNSET
        if not isinstance(self.sensitive_fields, Unset):
            sensitive_fields = self.sensitive_fields

        datapoint_count = self.datapoint_count
        app_link = self.app_link
        warning = self.warning

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "id": id,
                "project_id": project_id,
                "name": name,
                "tags": tags,
                "time_created": time_created,
            }
        )
        if models is not UNSET:
            field_dict["models"] = models
        if sensitive_fields is not UNSET:
            field_dict["sensitive_fields"] = sensitive_fields
        if datapoint_count is not UNSET:
            field_dict["datapoint_count"] = datapoint_count
        if app_link is not UNSET:
            field_dict["app_link"] = app_link
        if warning is not UNSET:
            field_dict["warning"] = warning

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.model_under_test_response_models import ModelUnderTestResponseModels

        d = src_dict.copy()
        id = d.pop("id")

        project_id = d.pop("project_id")

        name = d.pop("name")

        tags = cast(List[str], d.pop("tags"))

        time_created = d.pop("time_created")

        _models = d.pop("models", UNSET)
        models: Union[Unset, ModelUnderTestResponseModels]
        if isinstance(_models, Unset):
            models = UNSET
        else:
            models = ModelUnderTestResponseModels.from_dict(_models)

        sensitive_fields = cast(List[str], d.pop("sensitive_fields", UNSET))

        datapoint_count = d.pop("datapoint_count", UNSET)

        app_link = d.pop("app_link", UNSET)

        warning = d.pop("warning", UNSET)

        model_under_test_response = cls(
            id=id,
            project_id=project_id,
            name=name,
            tags=tags,
            time_created=time_created,
            models=models,
            sensitive_fields=sensitive_fields,
            datapoint_count=datapoint_count,
            app_link=app_link,
            warning=warning,
        )

        model_under_test_response.additional_properties = d
        return model_under_test_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import os
import shutil
from datetime import datetime

import pytest
from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.reporter import JSONReporter
from okareo_api_client.models.test_run_item import TestRunItem


@pytest.fixture(scope="module")
def rnd() -> str:
    return f'{random_string(5)} {datetime.now().strftime("%Y%m%d%H%M%S")}'


@pytest.fixture(scope="module")
def okareo_client() -> Okareo:
    return Okareo(api_key=API_KEY)


def test_new_jsonreporter(okareo_client: Okareo) -> None:
    json_data = {
        "id": "af197dd9-8d93-4ae3-9407-d0e19050bd44",
        "project_id": "5ff115a2-f4b0-4100-bbde-87a54428add4",
        "mut_id": "374b760a-ddf3-4295-aa41-57f6da65b67f",
        "scenario_set_id": "c6a6d412-734f-4799-8dc6-667aa75a8e30",
        "name": "CLI Evaluation-078355E56D - EVAL",
        "tags": [],
        "type": "NL_GENERATION",
        "start_time": "2024-03-19T16:27:37.171934",
        "end_time": "2024-03-19T16:28:06.531265",
        "test_data_point_count": 3,
        "model_metrics": {
            "mean_scores": {
                "coherence": 3.6041134314518946,
                "consistency": 3.6666003818872697,
                "fluency": 2.0248845922814245,
                "relevance": 2.3333201448723386,
                "overall": 2.907229637623232,
            },
            "scores_by_row": [
                {
                    "scenario_index": 1,
                    "coherence": 4.999831347314708,
                    "consistency": 4.999918368197127,
                    "fluency": 0,
                    "relevance": 4.998501009778189,
                    "overall": 3.749562681322506,
                    "test_id": "6b7cbe8d-c653-4dce-8de9-137a61566876",
                },
                {
                    "scenario_index": 2,
                    "coherence": 1,
                    "consistency": 1.000000023588648,
                    "fluency": 3,
                    "relevance": 1.0000000195556844,
                    "overall": 1.500000010786083,
                    "test_id": "d3dac533-334a-4c61-8495-08d7c82bcfcf",
                },
                {
                    "scenario_index": 3,
                    "coherence": 4.812508947040976,
                    "consistency": 4.999882753876036,
                    "fluency": 3.0746537768442743,
                    "relevance": 1.001459405283143,
                    "overall": 3.472126220761107,
                    "test_id": "bd40e350-5af3-4ccf-861c-d1a1889ed889",
                },
            ],
        },
        "error_matrix": [],
        "app_link": "https://app.okareo.com/project/5ff115a2-f4b0-4100-bbde-87a54428add4/eval/af197dd9-8d93-4ae3-9407-d0e19050bd44",
    }

    eval_item = TestRunItem.from_dict(json_data)
    reporter = JSONReporter([eval_item])
    assert reporter
    reporter.log()
    report_path = "_temp_reports_"
    os.environ["OKAREO_REPORT_DIR"] = report_path
    reporter.log()
    report_path_exists = os.path.exists(report_path)
    assert report_path_exists
    if os.path.isdir(report_path):
        shutil.rmtree(report_path)
    os.environ["OKAREO_REPORT_DIR"] = ""



import asyncio
import inspect
import json
import logging
import ssl
import threading
import urllib
from abc import abstractmethod
from base64 import b64encode
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import aiohttp
from attrs import define
from attrs import define as _attrs_define
from attrs import field
from nkeys import from_seed  # type: ignore
from tqdm import tqdm  # type: ignore

from okareo.error import MissingApiKeyError, MissingVectorDbError, TestRunError
from okareo_api_client.api.default import (
    add_datapoint_v0_datapoints_post,
    get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get,
    get_test_run_v0_test_runs_test_run_id_get,
    internal_custom_model_listener_v0_internal_custom_model_listener_get,
    run_test_v0_test_run_post,
    submit_test_v0_test_run_submit_post,
)
from okareo_api_client.client import Client
from okareo_api_client.errors import UnexpectedStatus
from okareo_api_client.models import (
    DatapointResponse,
    DatapointSchema,
    ModelUnderTestResponse,
    ScenarioDataPoinResponse,
    TestRunItem,
    TestRunType,
)
from okareo_api_client.models.error_response import ErrorResponse
from okareo_api_client.models.scenario_set_response import ScenarioSetResponse
from okareo_api_client.models.test_run_payload_v2 import TestRunPayloadV2
from okareo_api_client.models.test_run_payload_v2_api_keys import (
    TestRunPayloadV2ApiKeys,
)
from okareo_api_client.models.test_run_payload_v2_metrics_kwargs import (
    TestRunPayloadV2MetricsKwargs,
)
from okareo_api_client.models.test_run_payload_v2_model_results import (
    TestRunPayloadV2ModelResults,
)
from okareo_api_client.types import UNSET, Unset

from .async_utils import AsyncProcessorMixin

## BEGIN Monkey Patch for nats to use proxy env vars (via aiohttp)
#  Apply monkey patch at module level to allow aiohttp client session to pull proxy env vars
_original_client_session = aiohttp.ClientSession


# Create a wrapper class that inherits from the original ClientSession
class PatchedClientSession(_original_client_session):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        logging.debug("Applying proxy patch to aiohttp")
        logging.debug("Proxy env vars: %s", urllib.request.getproxies())
        kwargs["trust_env"] = True
        super().__init__(*args, **kwargs)


# Replace the original ClientSession with our patched version
aiohttp.ClientSession = PatchedClientSession  # type: ignore
# We import nats only after patching aiohttp to respect the proxy env vars
import nats  # type: ignore # noqa: E402

## END Monkey Patch for nats to use proxy env vars (via aiohttp)


class BaseModel:
    type: str
    api_key: Optional[str]

    @abstractmethod
    def params(self) -> dict:
        pass


class ModelUnderTest(AsyncProcessorMixin):
    """A class for managing a Model Under Test (MUT) in Okareo.
    Returned by [okareo.register_model()](/docs/reference/python-sdk/okareo#register_model)
    """

    def __init__(
        self,
        client: Client,
        api_key: str,
        mut: ModelUnderTestResponse,
        models: Optional[Dict[str, Any]] = None,
    ):
        self.client = client
        self.api_key = api_key

        self.mut_id = mut.id
        self.project_id = mut.project_id
        self.name = mut.name
        self.tags = mut.tags
        self.models = models
        self.app_link = mut.app_link
        super().__init__(name="OkareoDatapointsProcessor")

    def get_client(self) -> Client:
        return self.client

    def get_api_key(self) -> str:
        return self.api_key

    def add_data_point(
        self,
        input_obj: Union[dict, str, None] = None,
        result_obj: Union[dict, str, None] = None,
        feedback: Union[float, None] = None,
        context_token: Union[str, None] = None,
        error_message: Union[str, None] = None,
        error_code: Union[str, None] = None,
        input_datetime: Union[str, Unset] = UNSET,
        result_datetime: Union[str, Unset] = UNSET,
        project_id: Union[str, None] = None,
        tags: Union[List[str], None] = None,
        test_run_id: Union[None, str] = None,
        group_id: Union[None, str] = None,
    ) -> Union[DatapointResponse, ErrorResponse]:
        body = {
            "tags": tags or [],
            "input": json.dumps(input_obj, default=str),
            "result": json.dumps(result_obj, default=str),
            "context_token": context_token,
            "feedback": feedback,
            "error_message": error_message,
            "error_code": error_code,
            "input_datetime": (
                datetime.now().isoformat()
                if input_datetime == UNSET and input_obj is not None
                else input_datetime
            ),
            "result_datetime": (
                datetime.now().isoformat()
                if result_datetime == UNSET and result_obj is not None
                else result_datetime
            ),
            "project_id": self.project_id,
            "mut_id": self.mut_id,
            "test_run_id": test_run_id,
            "group_id": group_id,
        }
        response = add_datapoint_v0_datapoints_post.sync(
            client=self.client,
            api_key=self.api_key,
            json_body=DatapointSchema.from_dict(body),
        )
        if not response:
            print("Empty response from API")
        assert response is not None

        return response

    def add_data_point_async(
        self,
        input_obj: Union[dict, str, None] = None,
        result_obj: Union[dict, str, None] = None,
        feedback: Union[float, None] = None,
        context_token: Union[str, None] = None,
        error_message: Union[str, None] = None,
        error_code: Union[str, None] = None,
        input_datetime: Union[str, Unset] = UNSET,
        result_datetime: Union[str, Unset] = UNSET,
        project_id: Union[str, None] = None,
        tags: Union[List[str], None] = None,
        test_run_id: Union[None, str] = None,
        group_id: Union[None, str] = None,
    ) -> bool:
        body = {
            "tags": tags or [],
            "input": json.dumps(input_obj, default=str),
            "result": json.dumps(result_obj, default=str),
            "context_token": context_token,
            "feedback": feedback,
            "error_message": error_message,
            "error_code": error_code,
            "input_datetime": (
                datetime.now().isoformat()
                if input_datetime == UNSET and input_obj is not None
                else input_datetime
            ),
            "result_datetime": (
                datetime.now().isoformat()
                if result_datetime == UNSET and result_obj is not None
                else result_datetime
            ),
            "project_id": self.project_id,
            "mut_id": self.mut_id,
            "test_run_id": test_run_id,
            "group_id": group_id,
        }

        return self.async_call(
            add_datapoint_v0_datapoints_post.sync, DatapointSchema.from_dict(body)
        )

    def _validate_run_test_params(
        self,
        api_key: Optional[str],
        api_keys: Optional[dict],
        test_run_type: TestRunType,
    ) -> dict:
        assert isinstance(self.models, dict)
        model_names = list(self.models.keys())
        run_api_keys = api_keys if api_keys else {model_names[0]: api_key}

        if (
            "custom" not in model_names  # custom is a model without a key
            and "driver"
            not in model_names  # driver can have 2 keys for driver and target
            and len(model_names) != len(run_api_keys)
        ):
            raise MissingApiKeyError("Number of models and API keys does not match")

        if test_run_type == TestRunType.INFORMATION_RETRIEVAL:
            if {"pinecone", "qdrant", "custom"}.isdisjoint(model_names):
                raise MissingVectorDbError("No vector database specified")

        return run_api_keys

    def _has_custom_model(self) -> bool:
        assert isinstance(self.models, dict)
        if (
            "driver" in self.models
            and self.models["driver"]["target"]["type"] == "custom_target"
        ):
            return True
        custom_model_strs = ["custom", "custom_batch"]
        assert isinstance(self.models, dict)
        return any(
            model_str in list(self.models.keys()) for model_str in custom_model_strs
        )

    def _has_custom_batch_model(self) -> bool:
        assert isinstance(self.models, dict)
        return "custom_batch" in list(self.models.keys())

    def _get_scenario_data_points(
        self, scenario_id: str
    ) -> List[ScenarioDataPoinResponse]:
        scenario_data_points = (
            get_scenario_set_data_points_v0_scenario_data_points_scenario_id_get.sync(
                client=self.client,
                api_key=self.api_key,
                scenario_id=scenario_id,
            )
        )
        scenario_data_points = (
            scenario_data_points if isinstance(scenario_data_points, List) else []
        )
        return scenario_data_points

    def _add_model_invocation_for_scenario(
        self,
        custom_model_return_value: Any,
        model_data: dict,
        scenario_data_point_id: str,
    ) -> None:
        if isinstance(custom_model_return_value, ModelInvocation):
            model_prediction = custom_model_return_value.model_prediction
            model_output_metadata = custom_model_return_value.model_output_metadata
            model_input = custom_model_return_value.model_input
            tool_calls = custom_model_return_value.tool_calls
        else:  # assume the preexisting behavior of returning a tuple
            model_prediction, model_output_metadata = custom_model_return_value
            model_input = None
            tool_calls = None

        model_data["model_data"][scenario_data_point_id] = {
            "actual": model_prediction,
            "model_response": model_output_metadata,
            "model_input": model_input,
            "tool_calls": tool_calls,
        }

    def _get_test_run_payload(
        self,
        scenario_id: Any,
        name: str,
        api_key: Optional[str],
        api_keys: Optional[dict],
        run_api_keys: dict,
        metrics_kwargs: Optional[dict],
        test_run_type: TestRunType,
        calculate_metrics: bool,
        model_data: dict,
        checks: Optional[List[str]],
    ) -> TestRunPayloadV2:
        return TestRunPayloadV2(
            mut_id=self.mut_id,
            api_keys=(
                TestRunPayloadV2ApiKeys.from_dict(run_api_keys)
                if api_keys or api_key
                else UNSET
            ),
            scenario_id=scenario_id,
            name=name,
            type=test_run_type,
            calculate_metrics=calculate_metrics,
            metrics_kwargs=TestRunPayloadV2MetricsKwargs.from_dict(
                metrics_kwargs or {}
            ),
            model_results=(
                TestRunPayloadV2ModelResults.from_dict(model_data)
                if self._has_custom_model()
                else UNSET
            ),
            checks=checks if checks else UNSET,
        )

    async def connect_nats(self, user_jwt: str, seed: str, local_nats: str) -> Any:
        nc = None
        if local_nats != "":
            nc = await nats.connect(
                # Use the hostname 'nats' which is the service name in docker-compose
                servers=[local_nats],
                connect_timeout=120,
                allow_reconnect=True,
                max_reconnect_attempts=10,
                reconnect_time_wait=10,
            )
        else:
            nkey = from_seed(seed.encode())

            def user_jwt_cb() -> Any:
                return user_jwt.encode()

            def user_sign_cb(nonce: str) -> Any:
                sig = nkey.sign(nonce.encode())
                return b64encode(sig)

            ssl_ctx = ssl.create_default_context()
            ssl_ctx.check_hostname = True
            ssl_ctx.verify_mode = ssl.CERT_REQUIRED
            ngs_url = "wss://connect.ngs.global:443"
            nc = await nats.connect(
                servers=[ngs_url],
                user_jwt_cb=user_jwt_cb,
                signature_cb=user_sign_cb,
                tls=ssl_ctx,
                connect_timeout=30,
                allow_reconnect=True,
                max_reconnect_attempts=5,
                reconnect_time_wait=1,
            )
        return nc

    def call_custom_invoker(
        self,
        args: Any,
        message_history: Optional[list[dict[str, str]]] = None,
        scenario_input: Optional[Union[str, dict, list]] = None,
    ) -> Any:
        assert isinstance(self.models, dict)
        if self.models.get("custom"):
            return self.models["custom"]["model_invoker"](args)
        elif self.models.get("custom_batch"):
            return self.models["custom_batch"]["model_invoker"](args)
        else:
            messages = message_history if message_history is not None else args
            invoker = self.models["driver"]["target"]["model_invoker"]
            sig = inspect.signature(invoker)
            num_positional = sum(
                1
                for param in sig.parameters.values()
                if param.kind in (param.POSITIONAL_ONLY, param.POSITIONAL_OR_KEYWORD)
            )
            if num_positional > 1 and scenario_input is not None:
                # new impl; first pos arg is message_history, second is scenario_input
                return invoker(messages, scenario_input)
            else:
                # legacy impl; first pos arg is message_history (named args)
                return invoker(args)

    def get_params_from_custom_result(self, result: Any) -> Any:
        if isinstance(result, ModelInvocation):
            result = result.params()
        if (
            isinstance(result, list)
            and len(result) > 0
            and result[0].get("model_invocation")
        ):
            result = [r["model_invocation"].params() for r in result]
        return result

    async def _internal_run_custom_model_listener(
        self, stop_event: Any, nats_jwt: str, seed: str, local_nats: str
    ) -> None:
        nats_connection = await self.connect_nats(nats_jwt, seed, local_nats)
        try:

            async def message_handler_custom_model(msg: Any) -> None:
                try:
                    data = json.loads(msg.data.decode())
                    if data.get("close"):
                        await nats_connection.publish(
                            msg.reply, json.dumps({"status": "disconnected"}).encode()
                        )
                        stop_event.set()
                        return
                    args = data.get("args", [])
                    message_history = data.get("message_history", [])
                    scenario_input = data.get("scenario_input", None)
                    result = self.call_custom_invoker(
                        args, message_history, scenario_input
                    )
                    json_encodable_result = self.get_params_from_custom_result(result)
                    await nats_connection.publish(
                        msg.reply, json.dumps(json_encodable_result).encode()
                    )
                except Exception as e:
                    error_msg = f"An error occurred in the custom model invocation. {type(e).__name__}: {str(e)}"
                    print(error_msg)
                    await nats_connection.publish(
                        msg.reply, json.dumps({"error": error_msg}).encode()
                    )

            await nats_connection.subscribe(
                f"invoke.{self.mut_id}", cb=message_handler_custom_model
            )
            while not stop_event.is_set():
                await asyncio.sleep(0.1)
        except Exception as e:
            print(f"An error occurred in the custom model invocation: {str(e)}")
        finally:
            await nats_connection.close()

    def _internal_run_custom_model_thread(self, coro: Any) -> Any:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            return loop.run_until_complete(coro)
        finally:
            loop.close()

    def _internal_start_custom_model_thread(
        self, nats_jwt: str, seed: str, local_nats: str
    ) -> tuple:
        custom_model_thread_stop_event = threading.Event()
        custom_model_thread = threading.Thread(
            target=self._internal_run_custom_model_thread,
            args=(
                self._internal_run_custom_model_listener(
                    custom_model_thread_stop_event, nats_jwt, seed, local_nats
                ),
            ),
        )
        return custom_model_thread, custom_model_thread_stop_event

    def _internal_cleanup_custom_model(
        self,
        custom_model_thread_stop_event: threading.Event,
        custom_model_thread: threading.Thread,
    ) -> None:
        if custom_model_thread_stop_event:
            custom_model_thread_stop_event.set()  # Signal the thread to stop

        if custom_model_thread:
            custom_model_thread.join(timeout=5)

    def _run_test_internal(
        self,
        scenario: Union[ScenarioSetResponse, str],
        name: str,
        api_key: Optional[str] = None,
        api_keys: Optional[dict] = None,
        metrics_kwargs: Optional[dict] = None,
        test_run_type: TestRunType = TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics: bool = True,
        checks: Optional[List[str]] = None,
        run_test_method: Any = None,
    ) -> TestRunItem:
        """Internal method to run a test. This method is used by both run_test and submit_test."""
        self.custom_model_thread: Any = None
        self.custom_model_thread_stop_event: Any = None

        try:
            assert isinstance(self.models, dict)
            scenario_id = (
                scenario.scenario_id
                if isinstance(scenario, ScenarioSetResponse)
                else scenario
            )
            run_api_keys = self._validate_run_test_params(
                api_key, api_keys, test_run_type
            )

            model_data: dict = {"model_data": {}}
            if self._has_custom_model() and "driver" in self.models:
                creds = internal_custom_model_listener_v0_internal_custom_model_listener_get.sync(
                    client=self.client, api_key=self.api_key, mut_id=self.mut_id
                )
                assert isinstance(creds, dict)
                nats_jwt = creds["jwt"]
                seed = creds["seed"]
                local_nats = creds["local_nats"]
                (
                    self.custom_model_thread,
                    self.custom_model_thread_stop_event,
                ) = self._internal_start_custom_model_thread(nats_jwt, seed, local_nats)
                self.custom_model_thread.start()
            elif self._has_custom_model():
                self._custom_exec(scenario_id, model_data)

            response: TestRunItem = run_test_method(
                client=self.client,
                api_key=self.api_key,
                json_body=self._get_test_run_payload(
                    scenario_id,
                    name,
                    api_key,
                    api_keys,
                    run_api_keys,
                    metrics_kwargs,
                    test_run_type,
                    calculate_metrics,
                    model_data,
                    checks,
                ),
            )
            if isinstance(response, ErrorResponse):
                error_message = f"error: {response}, {response.detail}"
                print(error_message)
                raise TestRunError(str(response.detail))
            if not response:
                print("Empty response from API")
            assert response is not None
            return response
        except UnexpectedStatus as e:
            print(f"Unexpected status {e=}, {e.content=}")
            raise
        finally:
            self._internal_cleanup_custom_model(
                self.custom_model_thread_stop_event, self.custom_model_thread
            )

    def _check_multiturn_submit_safe(self, test_run_type: TestRunType) -> bool:
        """Check if the test_run_type is MULTI_TURN and if the model is a CustomMultiturnTarget.
        If so, return False to indicate that submit_test should not be used."""
        if (
            test_run_type == TestRunType.MULTI_TURN
            and self.models is not None
            and isinstance(self.models, dict)
            and "driver" in self.models
            and "target" in self.models["driver"]
            and "type" in self.models["driver"]["target"]
            and self.models["driver"]["target"]["type"] == "custom_target"
        ):
            return False
        return True

    def submit_test(
        self,
        scenario: Union[ScenarioSetResponse, str],
        name: str,
        api_key: Optional[str] = None,
        api_keys: Optional[dict] = None,
        metrics_kwargs: Optional[dict] = None,
        test_run_type: TestRunType = TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics: bool = True,
        checks: Optional[List[str]] = None,
    ) -> TestRunItem:
        """Asynchronous server-based version of test-run execution. For CustomModels, model
        invocations are handled client-side then evaluated server-side asynchronously. For other models,
        model invocations and evaluations handled server-side asynchronously.

        Arguments:
            scenario (Union[ScenarioSetResponse, str]): The scenario set or identifier to use for the test run.
            name (str): The name to assign to the test run.
            api_key (Optional[str]): Optional API key for authentication.
            api_keys (Optional[dict]): Optional dictionary of API keys for different services.
            metrics_kwargs (Optional[dict]): Optional dictionary of keyword arguments for metrics calculation.
            test_run_type (TestRunType): The type of test run to execute. Defaults to MULTI_CLASS_CLASSIFICATION.
            calculate_metrics (bool): Whether to calculate metrics after the test run. Defaults to True.
            checks (Optional[List[str]]): Optional list of checks to perform during the test run.

        Returns:
            TestRunItem: The resulting test run item for the submitted test run. The `id` field can be used to retrieve the test run.
        """
        endpoint = submit_test_v0_test_run_submit_post.sync
        if not self._check_multiturn_submit_safe(test_run_type):
            print(
                "WARNING: CustomMultiturnTarget models are not supported in submit_test. "
                + "Falling back to run_test instead."
            )
            endpoint = run_test_v0_test_run_post.sync
        return self._run_test_internal(
            scenario,
            name,
            api_key,
            api_keys,
            metrics_kwargs,
            test_run_type,
            calculate_metrics,
            checks,
            endpoint,
        )

    def run_test(
        self,
        scenario: Union[ScenarioSetResponse, str],
        name: str,
        api_key: Optional[str] = None,
        api_keys: Optional[dict] = None,
        metrics_kwargs: Optional[dict] = None,
        test_run_type: TestRunType = TestRunType.MULTI_CLASS_CLASSIFICATION,
        calculate_metrics: bool = True,
        checks: Optional[List[str]] = None,
    ) -> TestRunItem:
        """Server-based version of test-run execution. For CustomModels, model
        invocations are handled client-side then evaluated server-side. For other models,
        model invocations and evaluations handled server-side.

        Arguments:
            scenario (Union[ScenarioSetResponse, str]): The scenario set or identifier to use for the test run.
            name (str): The name to assign to the test run.
            api_key (Optional[str]): Optional API key for authentication.
            api_keys (Optional[dict]): Optional dictionary of API keys for different services.
            metrics_kwargs (Optional[dict]): Optional dictionary of keyword arguments for metrics calculation.
            test_run_type (TestRunType): The type of test run to execute. Defaults to MULTI_CLASS_CLASSIFICATION.
            calculate_metrics (bool): Whether to calculate metrics after the test run. Defaults to True.
            checks (Optional[List[str]]): Optional list of checks to perform during the test run.

        Returns:
            TestRunItem: The resulting test run item for the completed test run.
        """
        try:
            return self._run_test_internal(
                scenario,
                name,
                api_key,
                api_keys,
                metrics_kwargs,
                test_run_type,
                calculate_metrics,
                checks,
                run_test_v0_test_run_post.sync,
            )
        except Exception as e:
            raise TestRunError(str(e)) from e

    def get_test_run(self, test_run_id: str) -> TestRunItem:
        """Retrieve a test run by its ID.

        Arguments:
            test_run_id (str): The ID of the test run to retrieve.

        Returns:
            TestRunItem: The test run item corresponding to the provided ID.
        """
        try:
            response = get_test_run_v0_test_runs_test_run_id_get.sync(
                client=self.client, api_key=self.api_key, test_run_id=test_run_id
            )
            self.validate_response(response)
            assert isinstance(response, TestRunItem)

            return response
        except UnexpectedStatus as e:
            print(e.content)
            raise

    def validate_response(self, response: Any) -> None:
        if isinstance(response, ErrorResponse):
            error_message = f"error: {response}, {response.detail}"
            print(error_message)
            raise TypeError(error_message)
        if response is None:
            print("Received no response (None) from the API")
            raise ValueError("No response received")

    def _extract_input_from_scenario_data_point(
        self, scenario_data_point: ScenarioDataPoinResponse
    ) -> Any:
        """helper method to handle different scenario data point formats"""
        scenario_input = scenario_data_point.input_
        return scenario_input

    def _custom_exec(self, scenario_id: Any, model_data: Any) -> Any:
        assert isinstance(self.models, dict)

        assert scenario_id
        scenario_data_points = self._get_scenario_data_points(scenario_id)
        datapoint_len = len(scenario_data_points)

        if not self._has_custom_batch_model():
            custom_model_invoker = self.models["custom"]["model_invoker"]
            for scenario_data_point in tqdm(
                scenario_data_points, desc="Invoking CustomModel", unit="datapoint"
            ):
                scenario_input = self._extract_input_from_scenario_data_point(
                    scenario_data_point
                )

                custom_model_return_value = custom_model_invoker(scenario_input)
                self._add_model_invocation_for_scenario(
                    custom_model_return_value,
                    model_data,
                    scenario_data_point.id,
                )
        else:
            # batch inputs to the custom model
            custom_model_invoker = self.models["custom_batch"]["model_invoker"]
            batch_size = self.models["custom_batch"]["batch_size"]
            for index in tqdm(
                range(0, datapoint_len, batch_size),
                desc="Invoking CustomModel",
                unit="batch",
            ):
                end_index = min(index + batch_size, datapoint_len)
                scenario_data_points_batch = scenario_data_points[index:end_index]
                scenario_inputs = [
                    {
                        "id": sdp.id,
                        "input_value": self._extract_input_from_scenario_data_point(
                            sdp
                        ),
                    }
                    for sdp in scenario_data_points_batch
                ]
                custom_model_return_batch = custom_model_invoker(scenario_inputs)

                for return_dict in custom_model_return_batch:
                    self._add_model_invocation_for_scenario(
                        return_dict["model_invocation"],
                        model_data,
                        return_dict["id"],
                    )


@_attrs_define
class ModelInvocation:
    """
    Model invocation response object returned from a CustomModel.invoke method
    or as an element of a list returned from a CustomBatchModel.invoke_batch method.

    Arguments:
        model_prediction: Prediction from the model to be used when running the evaluation,
            e.g. predicted class from classification model or generated text completion from
            a generative model. This would typically be parsed out of the overall model_output_metadata.
        model_input: All the input sent to the model.
        model_output_metadata: Full model response, including any metadata returned with model's output.
        tool_calls: List of tool calls made during the model invocation, if any.
    """

    model_prediction: Union[dict, list, str, None] = None
    model_input: Union[dict, list, str, None] = None
    model_output_metadata: Union[dict, list, str, None] = None
    tool_calls: Optional[List] = None

    def params(self) -> dict:
        return {
            "actual": self.model_prediction,
            "model_input": self.model_input,
            "model_result": self.model_output_metadata,
            "tool_calls": self.tool_calls,
        }


@define
class OpenAIModel(BaseModel):
    """
    An OpenAI model definition with prompt template and relevant parameters for an Okareo evaluation.

    Arguments:
        model_id: Model ID to request from OpenAI completion.
            For list of available models, see https://platform.openai.com/docs/models
        temperature: Parameter for controlling the randomness of the model's output.
        system_prompt_template: `System` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`.
        user_prompt_template: `User` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`
        dialog_template: Dialog template in OpenAI message format to pass to the model. Uses mustache syntax for variable substitution.
        tools: List of tools to pass to the model.
    """

    type = "openai"
    model_id: str = field(default="gpt-4o-mini")
    temperature: float = field(default=0)
    system_prompt_template: Optional[str] = None
    user_prompt_template: Optional[str] = None
    dialog_template: Optional[str] = None
    tools: Optional[List] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "temperature": self.temperature,
            "system_prompt_template": self.system_prompt_template,
            "user_prompt_template": self.user_prompt_template,
            "dialog_template": self.dialog_template,
            "type": self.type,
            "tools": self.tools,
        }


@define
class GenerationModel(BaseModel):
    """An LLM definition with prompt template and relevant parameters for an Okareo evaluation.

    Arguments:
        model_id: Model ID to request for LLM completion.
        temperature: Parameter for controlling the randomness of the model's output.
        system_prompt_template: `System` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`.
        user_prompt_template: `User` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`
        dialog_template: Dialog template in OpenAI message format to pass to the model. Uses mustache syntax for variable substitution.
        tools: List of tools to pass to the model.
    """

    type = "generation"
    model_id: str = field(default="gpt-4o-mini")
    temperature: float = field(default=0)
    system_prompt_template: Optional[str] = None
    user_prompt_template: Optional[str] = None
    dialog_template: Optional[str] = None
    tools: Optional[List] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "temperature": self.temperature,
            "system_prompt_template": self.system_prompt_template,
            "user_prompt_template": self.user_prompt_template,
            "dialog_template": self.dialog_template,
            "type": self.type,
            "tools": self.tools,
        }


@_attrs_define
class OpenAIAssistantModel(BaseModel):
    """An OpenAI Assistant definition with prompt template and relevant parameters for an Okareo evaluation.

    Arguments:
        model_id: Assistant ID to request to run a thread against.
        assistant_prompt_template: `System` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`.
        user_prompt_template: `User` role prompt template to pass to the model. Uses mustache syntax for variable substitution, e.g. `{scenario_input}`
        dialog_template: Dialog template in OpenAI message format to pass to the model. Uses mustache syntax for variable substitution.
    """

    type = "openai_assistant"
    model_id: str
    assistant_prompt_template: Optional[str] = None
    user_prompt_template: Optional[str] = None
    dialog_template: Optional[str] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "assistant_prompt_template": self.assistant_prompt_template,
            "user_prompt_template": self.user_prompt_template,
            "dialog_template": self.dialog_template,
            "type": self.type,
        }


@_attrs_define
class CohereModel(BaseModel):
    """
    A Cohere model definition with prompt template and relevant parameters for an Okareo evaluation.

    Arguments:
        model_id: Model ID to request for the Cohere completion.
            For a full list of available models, see https://docs.cohere.com/v2/docs/models
        model_type: Type of application for the Cohere model. Currently, we support 'classify' and 'embed'.
        input_type: Input type for the Cohere embedding model.
            For more details, see https://docs.cohere.com/v2/docs/embeddings#the-input_type-parameter
    """

    type = "cohere"
    model_id: str
    model_type: str
    input_type: Optional[str] = None

    def params(self) -> dict:
        return {
            "model_id": self.model_id,
            "model_type": self.model_type,
            "input_type": self.input_type,
        }


@_attrs_define
class PineconeDb(BaseModel):
    """
    A Pinecone vector database configuration for use in an Okareo retrieval evaluation.

    Arguments:
        index_name: The name of the Pinecone index to connect to.
        region: The region where the Pinecone index is hosted.
        project_id: The project identifier associated with the Pinecone index.
        top_k: The number of top results to retrieve for queries. Defaults to 5.
    """

    type = "pinecone"
    index_name: str
    region: str
    project_id: str
    top_k: int = 5

    def params(self) -> dict:
        return {
            "index_name": self.index_name,
            "region": self.region,
            "project_id": self.project_id,
            "top_k": self.top_k,
        }


@_attrs_define
class QdrantDB(BaseModel):
    """
    A Qdrant vector database configuration for use in an Okareo retrieval evaluation.

    Arguments:
        collection_name: The name of the Qdrant collection to connect to.
        url: The URL of the Qdrant instance.
        top_k: The number of top results to retrieve for queries. Defaults to 5.
        sparse: Whether to use sparse vectors for the Qdrant collection. Defaults to False.
    """

    type = "qdrant"
    collection_name: str
    url: str
    top_k: int = 5
    sparse: bool = False

    def params(self) -> dict:
        return {
            "collection_name": self.collection_name,
            "url": self.url,
            "top_k": self.top_k,
            "sparse": self.sparse,
        }


@_attrs_define
class CustomModel(BaseModel):
    """A custom model definition for an Okareo evaluation.
    Requires a valid `invoke` definition that operates on a single input.

    Arguments:
        name: A name for the custom model.
    """

    type = "custom"
    name: str

    @abstractmethod
    def invoke(
        self, input_value: Union[dict, list, str]
    ) -> Union[ModelInvocation, Any]:
        """Method for taking a single scenario input and returning a single model output

        Arguments:
            input_value: Union[dict, list, str] - input to the model.

        Returns:
            Union[ModelInvocation, Any] - model output.
            If the model returns a ModelInvocation, it should contain the model's prediction, input, and metadata.
            If the model returns a tuple, the first element should be the model's prediction and the second element should be the metadata.
        """

    def params(self) -> dict:
        return {
            "name": self.name,
            "type": self.type,
            "model_invoker": self.invoke,
        }


@_attrs_define
class CustomMultiturnTarget(BaseModel):
    """A custom model definition for an Okareo multiturn evaluation.
    Requires a valid `invoke` definition that operates on a single turn of a converstation.
    """

    type = "custom_target"
    name: str

    @abstractmethod
    def invoke(
        self,
        messages: List[dict[str, str]],
        scenario_input: Optional[Union[dict, list, str]] = None,
    ) -> Union[ModelInvocation, Any]:
        """Method for continuing a multiturn conversation with a custom model

        Arguments:
            messages: list - list of messages in the conversation
            scenario_input: Optional[dict | list | str] - scenario input for the conversation

        Returns:
            Union[ModelInvocation, Any] - model output.
            If the model returns a ModelInvocation, it should contain the model's prediction, input, and metadata.
            If the model returns a tuple, the first element should be the model's prediction and the second element should be the metadata.
        """

    def params(self) -> dict:
        return {
            "name": self.name,
            "type": self.type,
            "model_invoker": self.invoke,
        }


@define
class StopConfig:
    """
    Configuration for stopping a multiturn conversation based on a specific check.

    Arguments:
        check_name: Name of the check to use for stopping the conversation.
        stop_on: The check condition to stop the conversation.
            Defaults to `True` (i.e., conversation stops when check evaluates to `True`).
    """

    check_name: str
    stop_on: bool = field(default=True)

    def params(self) -> dict:
        return {"check_name": self.check_name, "stop_on": self.stop_on}


class SessionConfig:
    """Configuration for a custom API endpoint that starts a session.

    Arguments:
        url: URL of the endpoint to start the session.
        method: HTTP method to use for the request. Defaults to `POST`.
        headers: Headers to include in the request. Defaults to an empty JSON object.
        body: Body to include in the request. Defaults to an empty JSON object.
        status_code: Expected HTTP status code of the response.
        response_session_id_path: Path to extract the session ID from the response.
            E.g., `response.id` will use the `id` field of the response JSON object to set the `session_id`.
    """

    def __init__(
        self,
        url: str,
        method: str = "POST",
        headers: Optional[Union[str, dict]] = None,
        body: Union[str, dict] = "{}",
        status_code: Optional[int] = None,
        response_session_id_path: str = "",
        response_message_path: str = "",
    ) -> None:
        self.url = url
        self.method = method
        self.headers = headers or json.dumps({})
        self.body = body
        self.status_code = status_code
        self.response_session_id_path = response_session_id_path
        self.response_message_path = response_message_path

    def to_dict(self) -> dict:
        return {
            "url": self.url,
            "method": self.method,
            "headers": self.headers,
            "body": self.body,
            "status_code": self.status_code,
            "response_session_id_path": self.response_session_id_path,
            "response_message_path": self.response_message_path,
        }


class TurnConfig:
    """
    Configuration for a custom API endpoint that continues a session/conversation by one turn.

    Arguments:
        url: URL of the endpoint to start the session.
        method: HTTP method to use for the request. Defaults to `POST`.
        headers: Headers to include in the request.
            Supports mustache syntax for variable substitution for `{latest_message}`, `{message_history}`, `{session_id}`.
            Defaults to an empty JSON object.
        body: Body to include in the request.
            Supports mustache syntax for variable substitution for `{latest_message}`, `{message_history}`, `{session_id}`.
            Defaults to an empty JSON object.
        status_code: Expected HTTP status code of the response.
        response_message_path: Path to extract the model's generated message from the response.
            E.g., `response.completion.message.content` will parse out the corresponding field of
            the response JSON object as the model's generated response.
        response_tool_calls_path: Path to extract tool calls from the response.
    """

    def __init__(
        self,
        url: str,
        method: str = "POST",
        headers: Optional[Union[str, dict]] = None,
        body: Union[str, dict] = "{}",
        status_code: Optional[int] = None,
        response_message_path: str = "",
        response_tool_calls_path: str = "",
    ) -> None:
        self.url = url
        self.method = method
        self.headers = headers or json.dumps({})
        self.body = body
        self.status_code = status_code
        self.response_message_path = response_message_path
        self.response_tool_calls_path = (
            response_tool_calls_path  # TODO: populate this once we support on BE
        )

    def to_dict(self) -> dict:
        return {
            "url": self.url,
            "method": self.method,
            "headers": self.headers,
            "body": self.body,
            "status_code": self.status_code,
            "response_message_path": self.response_message_path,
            "response_tool_calls_path": self.response_tool_calls_path,
        }


class EndSessionConfig:
    """Configuration for a custom API endpoint that ends a session.

    Arguments:
        url: URL of the endpoint to start the session.
        method: HTTP method to use for the request. Defaults to `POST`.
        headers: Headers to include in the request. Defaults to an empty JSON object.
        body: Body to include in the request. Defaults to an empty JSON object.
        status_code: Expected HTTP status code of the response.
        response_session_id_path: Path to extract the session ID from the response.
    """

    def __init__(
        self,
        url: str,
        method: str = "POST",
        headers: Optional[Union[str, dict]] = None,
        body: Union[str, dict] = "{}",
        status_code: Optional[int] = None,
        response_session_id_path: str = "",
    ) -> None:
        self.url = url
        self.method = method
        self.headers = headers or json.dumps({})
        self.body = body
        self.status_code = status_code
        self.response_session_id_path = response_session_id_path

    def to_dict(self) -> dict:
        return {
            "url": self.url,
            "method": self.method,
            "headers": self.headers,
            "body": self.body,
            "status_code": self.status_code,
        }


class CustomEndpointTarget:
    """
    A pair of custom API endpoints for starting a session and continuing a conversation to use in
    Okareo multiturn evaluation.

    Arguments:
        start_session: A valid SessionConfig for starting a session.
        next_turn: A valid TurnConfig for requesting and parsing the next turn of a conversation.
        end_session: A valid EndSessionConfig for ending a session.
        max_parallel_requests: Maximum number of parallel requests to allow when running the evaluation.
    """

    type = "custom_endpoint"

    def __init__(
        self,
        start_session: SessionConfig,
        next_turn: TurnConfig,
        end_session: Optional[EndSessionConfig] = None,
        max_parallel_requests: Optional[int] = None,
    ) -> None:
        self.start_session = start_session
        self.next_turn = next_turn
        self.end_session = end_session
        self.max_parallel_requests = max_parallel_requests

    def params(self) -> dict:
        return {
            "type": self.type,
            "start_session_params": self.start_session.to_dict(),
            "next_message_params": self.next_turn.to_dict(),
            "end_session_params": (
                self.end_session.to_dict() if self.end_session is not None else {}
            ),
            "max_parallel_requests": self.max_parallel_requests,
        }


@_attrs_define
class MultiTurnDriver(BaseModel):
    """
    A driver model for Okareo multiturn evaluation.

    Arguments:
        target: Target model under test to use in the multiturn evaluation.
        stop_check: A valid StopConfig or a dict that can be converted to StopConfig.
        driver_model_id: Model ID to use for the driver model (e.g., "gpt-4.1").
        driver_temperature: Parameter for controlling the randomness of the driver model's output.
        repeats: Number of times to run a conversation per scenario row. Defaults to 1.
        max_turns: Maximum number of turns to run in a conversation. Defaults to 5.
        first_turn: Name of model (i.e., "target" or "driver") that should initiate each conversation. Defaults to "target".
        driver_prompt_template: Optional system prompt template to pass to the driver model.
            Uses mustache syntax for variable substitution, e.g. `{input}`.
    """

    type = "driver"
    target: Union[
        OpenAIModel, CustomMultiturnTarget, GenerationModel, CustomEndpointTarget
    ]
    stop_check: Union[StopConfig, dict, None] = None
    driver_model_id: Optional[str] = None
    driver_temperature: Optional[float] = 0.8
    repeats: Optional[int] = 1
    max_turns: Optional[int] = 5
    first_turn: Optional[str] = "target"
    driver_prompt_template: Optional[str] = None
    checks_at_every_turn: Optional[bool] = False

    def __attrs_post_init__(self) -> None:
        if isinstance(self.stop_check, dict):
            self.stop_check = StopConfig(**self.stop_check)

    def params(self) -> dict:
        return {
            "type": self.type,
            "target": self.target.params(),
            "driver_model_id": self.driver_model_id,
            "driver_temperature": self.driver_temperature,
            "driver_prompt_template": self.driver_prompt_template,
            "repeats": self.repeats,
            "max_turns": self.max_turns,
            "first_turn": self.first_turn,
            "stop_check": (
                self.stop_check.params()
                if isinstance(self.stop_check, StopConfig)
                else self.stop_check
            ),
            "checks_at_every_turn": self.checks_at_every_turn,
        }


@_attrs_define
class CustomBatchModel(BaseModel):
    """A custom batch model definition for an Okareo evaluation.
    Requires a valid `invoke_batch` definition that operates on a single input.
    """

    type = "custom_batch"
    name: str
    batch_size: int = 1

    @abstractmethod
    def invoke_batch(
        self, input_batch: list[dict[str, Union[dict, list, str]]]
    ) -> list[dict[str, Union[ModelInvocation, Any]]]:
        """Method for taking a batch of scenario inputs and returning a corresponding batch of model outputs

        Arguments:
            input_batch: list[dict[str, Union[dict, list, str]]] - batch of inputs to the model. Expects a list of
            dicts of the format `{ 'id': str, 'input_value': Union[dict, list, str] }`.

        Returns:
            List of dicts of format `{ 'id': str, 'model_invocation': Union[ModelInvocation, Any] }`. 'id' must match
            the corresponding input_batch element's 'id'.
        """

    def params(self) -> dict:
        return {
            "name": self.name,
            "type": self.type,
            "batch_size": self.batch_size,
            "model_invoker": self.invoke_batch,
        }



import json
from typing import Any

from okareo_tests.common import API_KEY, random_string

from okareo import Okareo
from okareo.crewai_logger import CrewAILogger
from okareo_api_client.models.datapoint_search import DatapointSearch


def get_logger_config() -> dict[str, Any]:
    logger_config = {
        "api_key": API_KEY,
        "tags": ["crewai-test"],
        "context_token": random_string(10),
    }
    return logger_config


def test_crewai_logger() -> None:
    logger_config = get_logger_config()
    CrewAILogger(logger_config)

    from opentelemetry import trace

    tracer = trace.get_tracer("crewai.telemetry")
    with tracer.start_as_current_span("test_span") as span:
        span.set_attribute(
            "test_attribute",
            json.dumps(
                {
                    "agent": "test_agent",
                    "message": "This is a test message",
                    "timestamp": "2024-03-14T12:00:00Z",
                    "metadata": {"task_id": "123456", "priority": "high"},
                }
            ),
        )

    okareo = Okareo(api_key=API_KEY)
    dp = okareo.find_datapoints(
        DatapointSearch(context_token=logger_config["context_token"])
    )

    assert isinstance(dp, list)


def test_create_group() -> Any:
    okareo = Okareo(api_key=API_KEY)
    group_name = f"test_group_{random_string(5)}"
    tags = ["test", "crewai"]
    group = okareo.create_group(name=group_name, tags=tags)
    assert isinstance(group, dict)
    assert group.get("name") == group_name
    assert set(group.get("tags", [])) == set(tags)


def test_add_model_to_group() -> Any:
    okareo = Okareo(api_key=API_KEY)
    group_name = f"test_group_{random_string(5)}"
    group = okareo.create_group(name=group_name)
    model = okareo.register_model(name=f"test_model_{random_string(5)}")
    result = okareo.add_model_to_group(group, model)
    assert isinstance(result, dict)


def test_create_trace_eval() -> Any:
    okareo = Okareo(api_key=API_KEY)
    group_name = f"test_group_{random_string(5)}"
    group = okareo.create_group(name=group_name)
    okareo.create_trace_eval(group, "context_token")
